"""
This file defines the data types that are loaded from the data files.

"""
import numba as nb
import numpy as np

from oasislmf.pytools.common.data import areaperil_int, oasis_int, oasis_float
from oasislmf.pytools.common.event_stream import MAX_LOSS_IDX, CHANCE_OF_LOSS_IDX, TIV_IDX, STD_DEV_IDX, MEAN_IDX

items_data_type = nb.from_dtype(np.dtype([('item_id', oasis_int),
                                          ('damagecdf_i', oasis_int),
                                          ('rng_index', oasis_int)
                                          ]))

items_MC_data_type = nb.from_dtype(np.dtype([('item_id', oasis_int),
                                             ('areaperil_id', areaperil_int),
                                             ('vulnerability_id', oasis_int),
                                             ('hazcdf_i', oasis_int),
                                             ('rng_index', oasis_int),
                                             ('hazard_rng_index', oasis_int),
                                             ('eff_vuln_cdf_i', oasis_int),
                                             ('eff_vuln_cdf_Ndamage_bins', oasis_int)
                                             ]))

VulnCdfLookup = nb.from_dtype(np.dtype([('start', oasis_int), ('length', oasis_int)]))


coverage_type = nb.from_dtype(np.dtype([('tiv', np.float64),
                                        ('max_items', oasis_int),
                                        ('start_items', oasis_int),
                                        ('cur_items', oasis_int)
                                        ]))

NP_BASE_ARRAY_SIZE = 8

# Gul stream special sample idx
SPECIAL_SIDX = np.array([MAX_LOSS_IDX, CHANCE_OF_LOSS_IDX, TIV_IDX, STD_DEV_IDX, MEAN_IDX], dtype=oasis_int)
NUM_IDX = SPECIAL_SIDX.shape[0]

ITEM_MAP_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int32))
ITEM_MAP_VALUE_TYPE = nb.types.UniTuple(nb.types.int32, 3)
AREAPERIL_TO_EFF_VULN_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int64))
AREAPERIL_TO_EFF_VULN_VALUE_TYPE = nb.types.UniTuple(nb.types.int32, 2)

# compute the relative size of oasis_float and areaperil_int vs int32
oasis_float_to_int32_size = oasis_float.itemsize // np.int32().itemsize
areaperil_int_to_int32_size = areaperil_int.itemsize // np.int32().itemsize

haz_cdf_type = nb.from_dtype(np.dtype([('probability', oasis_float),
                                       ('intensity_bin_id', np.int32)]))

ProbMean = nb.from_dtype(np.dtype([('prob_to', oasis_float),
                                   ('bin_mean', oasis_float)
                                   ]))
ProbMean_size = ProbMean.size

damagecdfrec_stream = nb.from_dtype(np.dtype([('event_id', np.int32),
                                              ('areaperil_id', areaperil_int),
                                              ('vulnerability_id', np.int32)
                                              ]))

damagecdfrec = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                       ('vulnerability_id', np.int32)
                                       ]))


gulSampleslevelHeader = nb.from_dtype(np.dtype([('event_id', 'i4'),
                                                ('item_id', 'i4'),
                                                ]))
gulSampleslevelHeader_size = gulSampleslevelHeader.size

gulSampleslevelRec = nb.from_dtype(np.dtype([('sidx', 'i4'),
                                             ('loss', oasis_float),
                                             ]))
gulSampleslevelRec_size = gulSampleslevelRec.size


Keys = {'LocID': np.int32,
        'PerilID': 'category',
        'CoverageTypeID': np.int32,
        'AreaPerilID': areaperil_int,
        'VulnerabilityID': np.int32}



"""
This file defines the data types that are loaded from the data files.
"""
import numba as nb
import numpy as np

from oasislmf.pytools.common.data import areaperil_int, oasis_float

# Footprint file formats in order of priority
fp_format_priorities = ['parquet', 'binZ', 'bin', 'csv', 'parquet_dynamic']

# filenames
footprint_filename = 'footprint.bin'
footprint_index_filename = 'footprint.idx'
zfootprint_filename = 'footprint.bin.z'
zfootprint_index_filename = 'footprint.idx.z'
csvfootprint_filename = 'footprint.csv'
parquetfootprint_filename = 'footprint.parquet'
parquetfootprint_meta_filename = 'footprint_parquet_meta.json'
event_defintion_filename = 'event_definition.parquet'
hazard_case_filename = 'hazard_case.parquet'


FootprintHeader = nb.from_dtype(np.dtype([('num_intensity_bins', np.int32),
                                          ('has_intensity_uncertainty', np.int32)
                                          ]))

Event = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                ('intensity_bin_id', np.int32),
                                ('probability', oasis_float)
                                ]))

EventCSV = nb.from_dtype(np.dtype([('event_id', np.int32),
                                   ('areaperil_id', areaperil_int),
                                   ('intensity_bin_id', np.int32),
                                   ('probability', oasis_float)
                                   ]))

EventIndexBin = nb.from_dtype(np.dtype([('event_id', np.int32),
                                        ('offset', np.int64),
                                        ('size', np.int64)
                                        ]))

EventIndexBinZ = nb.from_dtype(np.dtype([('event_id', np.int32),
                                         ('offset', np.int64),
                                         ('size', np.int64),
                                         ('d_size', np.int64)
                                         ]))

Index_type = nb.from_dtype(np.dtype([('start', np.int64),
                                     ('end', np.int64)
                                     ]))

Vulnerability = nb.from_dtype(np.dtype([('vulnerability_id', np.int32),
                                        ('intensity_bin_id', np.int32),
                                        ('damage_bin_id', np.int32),
                                        ('probability', oasis_float)
                                        ]))

Item = nb.from_dtype(np.dtype([('id', np.int32),
                               ('coverage_id', np.int32),
                               ('areaperil_id', areaperil_int),
                               ('vulnerability_id', np.int32),
                               ('group_id', np.int32)
                               ]))

Event_defintion = nb.from_dtype(np.dtype([('section_id', np.int32),
                                          ('return_period', np.int32),
                                          ('rp_from', np.int32),
                                          ('rp_to', np.int32),
                                          ('interpolation', np.int32)
                                          ]))

Hazard_case = nb.from_dtype(np.dtype([('section_id', np.int32),
                                      ('areaperil_id', areaperil_int),
                                      ('return_period', np.int32),
                                      ('intensity', np.int32)
                                      ]))

Keys = {'LocID': np.int32,
        'PerilID': 'category',
        'CoverageTypeID': np.int32,
        'AreaPerilID': areaperil_int,
        'VulnerabilityID': np.int32}



__all__ = [
    'GenerateKeys',
    'GenerateKeysDeterministic'
]

import os
from packaging import version

from ods_tools.oed.setting_schema import AnalysisSettingSchema, ModelSettingSchema, OdsException
from ..base import ComputationStep
from ...lookup.factory import KeyServerFactory
from ...utils.exceptions import OasisException
from ...utils.coverages import SUPPORTED_COVERAGE_TYPES

from ...utils.inputs import str2bool
from ...utils.data import get_utctimestamp, get_exposure_data


class KeyComputationStep(ComputationStep):
    def get_exposure_data_config(self):
        return {
            'location': self.oed_location_csv,
            'oed_schema_info': self.oed_schema_info,
            'check_oed': self.check_oed,
            'use_field': True
        }


class GenerateKeys(KeyComputationStep):
    """
    Generates keys from a model lookup, and write Oasis keys and keys error files.

    The model lookup, which is normally independently implemented by the model
    supplier, should generate keys as dicts with the following format
    ::

        {
            "id": <loc. ID>,
            "peril_id": <OED sub-peril ID>,
            "coverage_type": <OED coverage type ID>,
            "area_peril_id": <area peril ID>,
            "vulnerability_id": <vulnerability ID>,
            "message": <loc. lookup status message>,
            "status": <loc. lookup status flag indicating success, failure or no-match>
        }

    The keys generation command can generate these dicts, and write them to
    file. It can also be used to write these to an Oasis keys file (which is a
    requirement for model execution), which has the following format.::

        LocID,PerilID,CoverageTypeID,AreaPerilID,VulnerabilityID
        ..
        ..
    This file only lists the locations for which there has been a successful
    lookup. The keys errors file lists all the locations with failing or
    non-matching lookups and has the following format::

        LocID,PerilID,CoverageTypeID,Message
        ..
        ..
    """

    step_params = [
        {'name': 'oed_location_csv', 'flag': '-x', 'is_path': True, 'pre_exist': True, 'help': 'Source location CSV file path'},
        {'name': 'oed_schema_info', 'is_path': True, 'pre_exist': True, 'help': 'path to custom oed_schema'},
        {'name': 'check_oed', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True, 'help': 'if True check input oed files'},
        {'name': 'keys_data_csv', 'flag': '-k', 'is_path': True, 'pre_exist': False, 'help': 'Generated keys CSV output path'},
        {'name': 'keys_errors_csv', 'flag': '-e', 'is_path': True, 'pre_exist': False, 'help': 'Generated keys errors CSV output path'},
        {'name': 'keys_format', 'flag': '-f', 'help': 'Keys files output format', 'choices': ['oasis', 'json'], 'default': 'oasis'},
        {'name': 'lookup_config_json', 'flag': '-g', 'is_path': True, 'pre_exist': False, 'help': 'Lookup config JSON file path'},
        {'name': 'lookup_data_dir', 'is_path': True, 'pre_exist': True, 'help': 'Model lookup/keys data directory path'},
        {'name': 'lookup_module_path', 'flag': '-l', 'is_path': True, 'pre_exist': False, 'help': 'Model lookup module path'},
        {'name': 'lookup_complex_config_json', 'flag': '-L', 'is_path': True, 'pre_exist': False, 'help': 'Complex lookup config JSON file path'},
        {'name': 'lookup_num_processes', 'type': int, 'default': -1, 'help': 'Number of workers in multiprocess pools'},
        {'name': 'lookup_num_chunks', 'type': int, 'default': -1, 'help': 'Number of chunks to split the location file into for multiprocessing'},
        {'name': 'model_version_csv', 'flag': '-v', 'is_path': True, 'pre_exist': False, 'help': 'Model version CSV file path'},
        {'name': 'model_settings_json', 'flag': '-M', 'is_path': True, 'pre_exist': True, 'help': 'Model settings JSON file path'},
        {'name': 'user_data_dir', 'flag': '-D', 'is_path': True, 'pre_exist': False,
         'help': 'Directory containing additional model data files which varies between analysis runs'},
        {'name': 'lookup_multiprocessing', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True,
         'help': 'Flag to enable/disable lookup multiprocessing'},
        {'name': 'disable_oed_version_update', 'type': str2bool, 'const': True, 'nargs': '?', 'default': False,
         'help': 'Flag to enable/disable conversion to latest compatible OED version. Must be present in model settings.'},

        # Manager only options
        {'name': 'verbose', 'default': False},
    ]

    def _get_output_dir(self):
        if self.keys_data_csv:
            return os.path.dirname(self.keys_data_csv)
        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        return os.path.join(os.getcwd(), 'runs', 'keys-{}'.format(utcnow))

    def run(self):
        if not (self.lookup_config_json or (self.lookup_data_dir and self.model_version_csv and self.lookup_module_path)):
            raise OasisException(
                'No pre-generated keys file provided, and no lookup assets '
                'provided to generate a keys file - if you do not have a '
                'pre-generated keys file then lookup assets must be provided - '
                'for a built-in lookup the lookup config. JSON file path must '
                'be provided, or for custom lookups the keys data path + model '
                'version file path + lookup package path must be provided'
            )

        output_dir = self._get_output_dir()
        output_type = 'json' if self.keys_format.lower() == 'json' else 'csv'
        if self.lookup_complex_config_json:
            AnalysisSettingSchema().validate_file(self.lookup_complex_config_json)

        exposure_data = get_exposure_data(self, add_internal_col=True)

        if not self.disable_oed_version_update:
            if self.model_settings_json is not None:
                try:
                    model_settings = ModelSettingSchema().get(self.model_settings_json, validate=False)
                    # Check for the existence and contents of 'supported_oed_versions'
                    supported_versions = model_settings.get('data_settings', {}).get('supported_oed_versions', None)
                    if supported_versions:
                        if isinstance(supported_versions, str):
                            self.logger.info(f"Converting to OED version {supported_versions}")
                            exposure_data.to_version(supported_versions)
                        elif isinstance(supported_versions, list) and supported_versions:
                            # If 'supported_oed_versions' is a list and is not empty
                            # Sort the versions in descending order
                            supported_versions = sorted(supported_versions, key=version.parse, reverse=True)
                            self.logger.info(f"Converting to OED version {supported_versions[0]}")
                            exposure_data.to_version(supported_versions[0])
                        else:
                            # If 'supported_oed_versions' is neither a string nor a non-empty list
                            self.logger.warning("Invalid OED version information in model settings.")
                    else:
                        # If 'supported_oed_versions' is missing or empty
                        self.logger.debug("No OED version information in model settings.")
                except OdsException:
                    self.logger.debug("No OED version information in model settings.")

        keys_fp = self.keys_data_csv or os.path.join(output_dir, f'keys.{output_type}')
        keys_errors_fp = self.keys_errors_csv or os.path.join(output_dir, f'keys-errors.{output_type}')
        os.makedirs(os.path.dirname(keys_fp), exist_ok=True)
        os.makedirs(os.path.dirname(keys_errors_fp), exist_ok=True)
        keys_success_msg = True if self.lookup_complex_config_json else False

        model_info, key_server = KeyServerFactory.create(
            lookup_config_fp=self.lookup_config_json,
            model_keys_data_path=self.lookup_data_dir,
            model_version_file_path=self.model_version_csv,
            lookup_module_path=self.lookup_module_path,
            complex_lookup_config_fp=self.lookup_complex_config_json,
            user_data_dir=self.user_data_dir,
            output_directory=output_dir
        )

        res = key_server.generate_key_files(
            location_df=exposure_data.location.dataframe,
            successes_fp=keys_fp,
            errors_fp=keys_errors_fp,
            format=self.keys_format,
            keys_success_msg=keys_success_msg,
            multiproc_enabled=self.lookup_multiprocessing,
            multiproc_num_cores=self.lookup_num_processes,
            multiproc_num_partitions=self.lookup_num_chunks,
        )

        self.logger.debug(f"key generated used model {model_info}")
        self.logger.info('\nKeys successful: {} generated with {} items'.format(res[0], res[1]))
        if len(res) == 4:
            self.logger.info('Keys errors: {} generated with {} items'.format(res[2], res[3]))
        return res


class GenerateKeysDeterministic(KeyComputationStep):
    step_params = [
        {'name': 'oed_location_csv', 'flag': '-x', 'is_path': True, 'pre_exist': True, 'help': 'Source location CSV file path'},
        {'name': 'oed_schema_info', 'is_path': True, 'pre_exist': True, 'help': 'path to custom oed_schema'},
        {'name': 'check_oed', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True, 'help': 'if True check input oed files'},
        {'name': 'keys_data_csv', 'flag': '-k', 'is_path': True, 'pre_exist': False, 'help': 'Generated keys CSV output path'},
        {'name': 'supported_oed_coverage_types', 'type': int, 'nargs': '+', 'default': list(v['id'] for v in SUPPORTED_COVERAGE_TYPES.values()),
         'help': 'Select List of supported coverage_types [1, .. ,4]'},
        {'name': 'model_perils_covered', 'nargs': '+', 'default': ['AA1'],
         'help': 'List of peril covered by the model'}
    ]

    def _get_output_dir(self):
        if self.keys_data_csv:
            return os.path.basename(self.keys_data_csv)
        utcnow = get_utctimestamp(fmt='%Y%m%d%H%M%S')
        return os.path.join(os.getcwd(), 'runs', 'keys-{}'.format(utcnow))

    def run(self):
        output_dir = self._get_output_dir()
        keys_fp = self.keys_data_csv or os.path.join(output_dir, 'keys.csv')
        os.makedirs(os.path.dirname(keys_fp), exist_ok=True)

        exposure_data = get_exposure_data(self, add_internal_col=True)
        config = {'builtin_lookup_type': 'peril_covered_deterministic',
                  'model': {"supplier_id": "OasisLMF",
                            "model_id": "Deterministic",
                            "model_version": "1"},
                  'supported_oed_coverage_types': self.supported_oed_coverage_types,
                  'model_perils_covered': self.model_perils_covered}

        model_info, lookup = KeyServerFactory.create(
            lookup_config=config,
            output_directory=output_dir
        )
        return lookup.generate_key_files(
            location_df=exposure_data.location.dataframe,
            successes_fp=keys_fp,
            format='oasis',
        )



__all__ = [
    'GenerateExposurePreAnalysisCmd',
    'GenerateKeysCmd',
    'GeneratePostFileGenCmd',
    'GeneratePrelossCmd',
    'GenerateLossesCmd',
    'GenerateLossesPartialCmd',
    'GenerateLossesOutputCmd',
    'GenerateOasisFilesCmd',
    'ModelCmd',
    'RunCmd'
]

from argparse import RawDescriptionHelpFormatter

from .command import OasisBaseCommand, OasisComputationCommand


class GenerateExposurePreAnalysisCmd(OasisComputationCommand):
    """
    Generate a new EOD from original one by specifying a model specific pre-analysis hook for exposure modification
    see ExposurePreAnalysis for more detail
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'ExposurePreAnalysis'


class GeneratePostFileGenCmd(OasisComputationCommand):
    """
    Generate a new EOD from original one by specifying a model specific pre-analysis hook for exposure modification
    see ExposurePreAnalysis for more detail
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PostFileGen'


class GeneratePrelossCmd(OasisComputationCommand):
    """
    Generate a new EOD from original one by specifying a model specific pre-analysis hook for exposure modification
    see ExposurePreAnalysis for more detail
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PreLoss'


class GenerateKeysCmd(OasisComputationCommand):
    """
    Generates keys from a model lookup, and write Oasis keys and keys error files.
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'GenerateKeys'


class GenerateOasisFilesCmd(OasisComputationCommand):
    """
    Generates the standard Oasis GUL input files + optionally the IL/FM input
    files and the RI input files.
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'GenerateOasisFiles'


class GenerateLossesCmd(OasisComputationCommand):
    """
    Generates the standard Oasis GUL input files + optionally the IL/FM input
    files and the RI input files.
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'GenerateOasisLosses'


class GenerateLossesPartialCmd(OasisComputationCommand):
    """
    Distributed Oasis CMD: desc todo
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'GenerateLossesPartial'


class GenerateLossesOutputCmd(OasisComputationCommand):
    """
    Distributed Oasis CMD: desc todo
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'GenerateLossesOutput'


class RunCmd(OasisComputationCommand):
    """
    Run models end to end.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'RunModel'


class RunPostAnalysisCmd(OasisComputationCommand):
    """
    Run the output postprocessing step.

    The command line arguments can be supplied in the configuration file
    (``oasislmf.json`` by default or specified with the ``--config`` flag).
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PostAnalysis'


class ModelCmd(OasisBaseCommand):
    """
    Model subcommands::

        * generating keys files from model lookups
        * generating Oasis input CSV files (GUL [+ IL, RI])
        * generating losses from a preexisting set of Oasis input CSV files
        * generating deterministic losses (no model)
        * running a model end-to-end
    """
    sub_commands = {
        'generate-exposure-pre-analysis': GenerateExposurePreAnalysisCmd,
        'generate-keys': GenerateKeysCmd,
        'generate-oasis-files': GenerateOasisFilesCmd,
        'generate-post-file-gen': GeneratePostFileGenCmd,
        'generate-pre-loss': GeneratePrelossCmd,
        'generate-losses': GenerateLossesCmd,
        'generate-losses-chunk': GenerateLossesPartialCmd,
        'generate-losses-output': GenerateLossesOutputCmd,
        'run': RunCmd,
        'run-postanalysis': RunPostAnalysisCmd,
    }



__all__ = [
    'load_oed_dfs',
    'OedValidator'
]
import os

from ods_tools.oed import fill_empty

from ..utils.data import get_dataframe, RI_INFO_DEFAULTS, RI_SCOPE_DEFAULTS


# TODO - add validator
class OedValidator(object):

    def __init__(self, ri_info_rules=None, ri_scope_rules=None):
        self.rules_ode_scope = ri_info_rules
        self.rules_ode_info = ri_scope_rules

        self.ri_info_required_cols = list(RI_INFO_REQUIRED_COLS)

        self.ri_info_defaults = {
            'CededPercent': 1.0,
            'RiskLimit': 0.0,
            'RiskAttachment': 0.0,
            'OccLimit': 0.0,
            'OccAttachment': 0.0,
            'TreatyShare': 0.0
        }

        self.ri_scope_required_cols = list(RI_SCOPE_REQUIRED_COLS)

        self.error_structure = {}

    def _unique_reins(self, reins_info_df):
        '''
        check if only one reins type exisits
        '''
        return (len(reins_info_df.ReinsType.unique()) == 1)

    def _find_missing(self, df_src, column_name, df_dest):
        '''
        return list of values if df_dest[`column_name`] values do not exist in df_src
        '''
        src_values = df_dest[column_name].unique().tolist()
        missing_df = df_src[~df_src.isin({column_name: src_values})].dropna()
        return missing_df[column_name].tolist()

    def _all_scope_non_specific(self, scope_df):
        return scope_df[['AccountNumber',
                         'PolicyNumber',
                         'LocationNumber'
                         ]].isnull().all().all()

    def _all_scope_specific(self, scope_df):
        return scope_df[['AccountNumber',
                         'PolicyNumber',
                         'LocationNumber'
                         ]].notnull().all().all()

    def _error_struture(self, check_type, chcek_scope, msg, info=''):
        return {
            "check": check_type,
            "scope": chcek_scope,
            "messsages": msg,
            "meta_data": info,
        }

    # def validate(self, account_df, location_df, ri_info_df, ri_scope_df):
    def validate(self, ri_info_df, ri_scope_df):
        '''
        Validate OED resinurance structure before running calculations.
        '''
        error_list = []

        for inuring_priority in range(1, ri_info_df['InuringPriority'].max() + 1):
            inuring_priority_ri_info_df = ri_info_df[ri_info_df.InuringPriority == inuring_priority]
            if inuring_priority_ri_info_df.empty:
                continue

            inuring_scope_ids = inuring_priority_ri_info_df.ReinsNumber.tolist()
            inuring_scopes = [ri_scope_df[ri_scope_df.ReinsNumber == ID] for ID in inuring_scope_ids]
            reins_types_found = inuring_priority_ri_info_df.ReinsType.unique().tolist()

            meta_data = {
                "InuringPriority": inuring_priority,
                "ReinsTypes": reins_types_found,
                "ri_info_ReinsNumbers": inuring_priority_ri_info_df.ReinsNumber.tolist(),
                "ri_info_line_nums": [idx + 2 for idx in inuring_priority_ri_info_df.index.tolist()],
            }

            # CHECK - only single ri_type is set per inuring priority
            if len(reins_types_found) > 1:
                error_list.append(self._error_struture(
                    "inuring_reins_type",
                    "RI_{}".format(inuring_priority),
                    "Inuring layer must have a unique ReinsType.",
                    meta_data
                ))
                continue
            elif len(reins_types_found) < 1:
                error_list.append(self._error_struture(
                    "inuring_reins_type",
                    "RI_{}".format(inuring_priority),
                    "Inuring layer missing a ReinsType.",
                    meta_data
                ))
                continue

            # CHECK - ri_type is supported
            ri_type = reins_types_found[0]
            if ri_type not in REINS_TYPES:
                error_list.append(self._error_struture(
                    "inuring_reins_type",
                    "RI_{}".format(inuring_priority),
                    "Unsupported ReinsType",
                    meta_data
                ))

            # CHECK scope of inuring layer
            for scope_df in inuring_scopes:
                scope_risk_levels = scope_df.RiskLevel.unique()
                meta_data.update({
                    "RiskLevels": scope_risk_levels.tolist(),
                    "ri_scope_ReinsNumber": scope_df.ReinsNumber.tolist(),
                    "ri_scope_line_nums": [idx + 2 for idx in scope_df.index.tolist()],
                })
                # CHECK - each scope only has one risk level type
                if len(scope_risk_levels) > 1:
                    error_list.append(self._error_struture(
                        "inuring_risk_level",
                        "RI_{}".format(inuring_priority),
                        "Mix of risk levels in a single reinsurance scope",
                        meta_data,
                    ))
                    continue
                elif len(scope_risk_levels) < 1:
                    error_list.append(self._error_struture(
                        "inuring_risk_level",
                        "RI_{}".format(inuring_priority),
                        "inuring layer has no reinsurance scope",
                        meta_data,
                    ))
                    continue

                # CHECK - Risk level is supported
                risk_level_id = scope_risk_levels[0]
                if risk_level_id not in SUPPORTED_RISK_LEVELS[ri_type]:
                    error_list.append(self._error_struture(
                        "inuring_risk_level",
                        "RI_{}".format(inuring_priority),
                        "Unsupported risk level",
                        meta_data,
                    ))

        return (not error_list, error_list)


def load_oed_dfs(oed_ri_info_file, oed_ri_scope_file, show_all=False):
    """
    Load OED data files.
    """
    do_reinsurance = True

    oed_ri_info_file_exists = os.path.exists(oed_ri_info_file)
    oed_ri_scope_file_exists = os.path.exists(oed_ri_scope_file)

    if not oed_ri_info_file_exists and not oed_ri_scope_file_exists:
        ri_info_df = None
        ri_scope_df = None
        do_reinsurance = False
    elif oed_ri_info_file_exists and oed_ri_scope_file_exists:
        ri_info_df = get_dataframe(
            oed_ri_info_file, lowercase_cols=False,
            required_cols=RI_INFO_REQUIRED_COLS,
            col_defaults=RI_INFO_DEFAULTS,
            col_dtypes=RI_INFO_DTYPES)
        ri_scope_df = get_dataframe(
            oed_ri_scope_file, lowercase_cols=False,
            required_cols=RI_SCOPE_REQUIRED_COLS,
            col_defaults=RI_SCOPE_DEFAULTS,
            col_dtypes=RI_SCOPE_DTYPES)

        # Treat empty Risk Level as portfolio level scope.
        # Also need nan, as this is produced when
        # a single row with empty Risk Level is loaded.
        fill_empty(ri_info_df, 'RiskLevel', REINS_RISK_LEVEL_PORTFOLIO)
    else:
        print("Both reinsurance files must exist: {} {}".format(
            oed_ri_info_file, oed_ri_scope_file))

    if do_reinsurance:
        optional_currency_cols = list({'originalcurrency', 'rateofexchange'} & set(ri_info_df.columns))
        ri_info_df = ri_info_df[OED_REINS_INFO_FIELDS + optional_currency_cols].copy()
        ri_scope_df = ri_scope_df[OED_REINS_SCOPE_FIELDS].copy()

        # Ensure Percent feilds are float
        info_float_cols = ['CededPercent', 'PlacedPercent', 'TreatyShare']
        scope_float_cols = ['CededPercent']
        ri_info_df[info_float_cols] = ri_info_df[info_float_cols].astype(float)
        ri_scope_df[scope_float_cols] = ri_scope_df[scope_float_cols].astype(float)

    return (ri_info_df, ri_scope_df, do_reinsurance)

#
# Ktools constants
#


DEDUCTIBLE_AND_LIMIT_CALCRULE_ID = 1
FRANCHISE_DEDUCTIBLE_AND_LIMIT_CALCRULE_ID = 3
DEDUCTIBLE_ONLY_CALCRULE_ID = 12
DEDUCTIBLE_AS_A_CAP_ON_THE_RETENTION_OF_INPUT_LOSSES_CALCRULE_ID = 10
DEDUCTIBLE_AS_A_FLOOR_ON_THE_RETENTION_OF_INPUT_LOSSES_CALCRULE_ID = 11
DEDUCTIBLE_LIMIT_AND_SHARE_CALCRULE_ID = 2
DEDUCTIBLE_AND_LIMIT_AS_A_PROPORTION_OF_LOSS_CALCRUKE_ID = 5
DEDUCTIBLE_WITH_LIMIT_AS_A_PROPORTION_OF_LOSS_CALCRUKE_ID = 9
LIMIT_ONLY_CALCRULE_ID = 14
LIMIT_AS_A_PROPORTION_OF_LOSS_CALCRULE_ID = 15
DEDUCTIBLE_AS_A_PROPORTION_OF_LOSS_CALCRULE_ID = 16

CALCRULE_ID_DEDUCTIBLE_AND_LIMIT = 1
CALCRULE_ID_DEDUCTIBLE_ATTACHMENT_LIMIT_AND_SHARE = 2
CALCRULE_ID_FRANCHISE_DEDUCTIBLE_AND_LIMIT = 3
CALCRULE_ID_DEDUCTIBLE_AND_LIMIT_PERCENT_TIV = 4
CALCRULE_ID_DEDUCTIBLE_AND_LIMIT_PERCENT_LOSS = 5
CALCRULE_ID_DEDUCTIBLE_PERCENT_TIV = 6
CALCRULE_ID_LIMIT_AND_MAX_DEDUCTIBLE = 7
CALCRULE_ID_LIMIT_AND_MIN_DEDUCTIBLE = 8
CALCRULE_ID_LIMIT_WITH_DEDUCTIBLE_PERCENT_LIMIT = 9
CALCRULE_ID_MAX_DEDUCTIBLE = 10
CALCRULE_ID_MIN_DEDUCTIBLE = 11
CALCRULE_ID_DEDUCTIBLE_ONLY = 12
CALCRULE_ID_MAIN_AND_MAX_DEDUCTIBLE = 13
CALCRULE_ID_LIMIT_ONLY = 14
CALCRULE_ID_LIMIT_PERCENT_LOSS = 15
CALCRULE_ID_DEDUCTIBLE_PERCENT_LOSS = 16
CALCRULE_ID_DEDUCTIBLE_PERCENT_LOSS_ATTACHMENT_LIMIT_AND_SHARE = 17
CALCRULE_ID_DEDUCTIBLE_PERCENT_TIV_ATTACHMENT_LIMIT_AND_SHARE = 18
CALCRULE_ID_DEDUCTIBLE_PERCENT_LOSS_WITH_MIN_AND_MAX = 19
CALCRULE_ID_REVERSE_FRANCHISE_DEDUCTIBLE = 20
CALCRULE_ID_SHARE_AND_LIMIT = 21
CALCRULE_ID_QUOTA_SHARE = 22
CALCRULE_ID_OCCURRENCE_LIMIT_AND_SHARE = 23
CALCRULE_ID_OCCURRENCE_CATASTROPHE_EXCESS_OF_LOSS = 24
CALCRULE_ID_FACULTATIVE_WITH_POLICY_SHARE = 25

NO_ALLOCATION_ALLOC_ID = 0
ALLOCATE_TO_ITEMS_BY_GUL_ALLOC_ID = 1
ALLOCATE_TO_ITEMS_BY_PREVIOUS_LEVEL_ALLOC_ID = 2

BUILDING_COVERAGE_TYPE_ID = 1
OTHER_BUILDING_COVERAGE_TYPE_ID = 2
CONTENTS_COVERAGE_TYPE_ID = 3
TIME_COVERAGE_TYPE_ID = 4
COVERAGE_TYPES = [
    BUILDING_COVERAGE_TYPE_ID,
    OTHER_BUILDING_COVERAGE_TYPE_ID,
    CONTENTS_COVERAGE_TYPE_ID,
    TIME_COVERAGE_TYPE_ID]

PERIL_WIND = 1
PERILS = [PERIL_WIND]

GUL_INPUTS_FILES = [
    'coverages',
    'gulsummaryxref',
    'items']

IL_INPUTS_FILES = [
    'fm_policytc',
    'fm_profile',
    'fm_programme',
    'fm_xref',
    'fmsummaryxref']

OPTIONAL_INPUTS_FILES = [
    'events']

CONVERSION_TOOLS = {
    'coverages': 'coveragetobin',
    'events': 'evetobin',
    'fm_policytc': 'fmpolicytctobin',
    'fm_profile': 'fmprofiletobin',
    'fm_programme': 'fmprogrammetobin',
    'fm_xref': 'fmxreftobin',
    'fmsummaryxref': 'fmsummaryxreftobin',
    'gulsummaryxref': 'gulsummaryxreftobin',
    'items': "itemtobin"}


NOT_SET_ID = -1
LARGE_VALUE = 9999999999999


# --- OED constants --------------------------------------------------------- #
RI_INFO_REQUIRED_COLS = [
    'ReinsNumber',
    'PlacedPercent',
    'InuringPriority',
    'ReinsType',
    'RiskLevel',
    'ReinsCurrency'
]

RI_SCOPE_REQUIRED_COLS = {
    'ReinsNumber',
}

RI_INFO_DTYPES = {
    'ReinsNumber': "int",
    'CededPercent': "float",
    'RiskLimit': "float",
    'RiskAttachment': "float",
    'OccLimit': "float",
    'OccAttachment': "float",
    'InuringPriority': "int",
    'ReinsType': "str",
    'RiskLevel': "str",
    'PlacedPercent': "float",
    'TreatyShare': "float",
    'ReinsCurrency': 'str'
}

RI_SCOPE_DTYPES = {
    'ReinsNumber': "int",
    'PortNumber': "str",
    'AccNumber': "str",
    'LocGroup': "str",
    'PolNumber': "str",
    'LocNumber': "str",
    'CedantName': "str",
    'ProducerName': "str",
    'LOB': "str",
    'CountryCode': "str",
    'ReinsTag': "str",
    'CededPercent': "float"
}


POLICYITEM_LEVEL = 0
LOCATION_LEVEL = 1
POLICY_LEVEL = 2
ACCOUNT_LEVEL = 3
OCCURRENCE_LEVEL = 4

PASSTHROUGH_NODE_TYPE = 1
NOLOSS_NODE_TYPE = 1

REINS_TYPE_FAC = "FAC"
REINS_TYPE_QUOTA_SHARE = "QS"
REINS_TYPE_SURPLUS_SHARE = "SS"
REINS_TYPE_PER_RISK = "PR"
REINS_TYPE_CAT_XL = "CXL"
REINS_TYPE_AGG_XL = "AXL"
REINS_TYPES = [
    REINS_TYPE_FAC,
    REINS_TYPE_QUOTA_SHARE,
    REINS_TYPE_SURPLUS_SHARE,
    REINS_TYPE_PER_RISK,
    REINS_TYPE_CAT_XL
]

REINS_RISK_LEVEL_PORTFOLIO = "SEL"
REINS_RISK_LEVEL_LOCATION = "LOC"
REINS_RISK_LEVEL_LOCATION_GROUP = "LGR"
REINS_RISK_LEVEL_POLICY = "POL"
REINS_RISK_LEVEL_ACCOUNT = "ACC"
REINS_RISK_LEVELS = [
    REINS_RISK_LEVEL_LOCATION,
    REINS_RISK_LEVEL_LOCATION_GROUP,
    REINS_RISK_LEVEL_POLICY,
    REINS_RISK_LEVEL_ACCOUNT,
    REINS_RISK_LEVEL_PORTFOLIO,
]


SUPPORTED_RISK_LEVELS = {
    REINS_TYPE_FAC: [REINS_RISK_LEVEL_LOCATION, REINS_RISK_LEVEL_LOCATION_GROUP, REINS_RISK_LEVEL_POLICY, REINS_RISK_LEVEL_ACCOUNT],
    REINS_TYPE_SURPLUS_SHARE: [REINS_RISK_LEVEL_LOCATION, REINS_RISK_LEVEL_LOCATION_GROUP, REINS_RISK_LEVEL_POLICY, REINS_RISK_LEVEL_ACCOUNT],
    REINS_TYPE_PER_RISK: [REINS_RISK_LEVEL_LOCATION, REINS_RISK_LEVEL_LOCATION_GROUP, REINS_RISK_LEVEL_POLICY, REINS_RISK_LEVEL_ACCOUNT],
    REINS_TYPE_CAT_XL: [REINS_RISK_LEVEL_PORTFOLIO],
    REINS_TYPE_QUOTA_SHARE: REINS_RISK_LEVELS
}

# Subset of the fields that are currently used
OED_ACCOUNT_FIELDS = [
    'PortNumber',
    'AccNumber',
    'PolNumber',
    'AccPeril',
    'AccDed6All',
    'AccLimit6All'
]

OED_LOCATION_FIELDS = [
    'AccNumber',
    'LocGroup',
    'LocNumber',
    'LocDed6All',
    'LocLimit6All',
    'BuildingTIV',
    'OtherTIV',
    'ContentsTIV',
    'BITIV'
]

OED_REINS_INFO_FIELDS = [
    'ReinsNumber',
    'CededPercent',
    'RiskLimit',
    'RiskAttachment',
    'OccLimit',
    'OccAttachment',
    'InuringPriority',
    'ReinsType',
    'RiskLevel',
    'PlacedPercent',
    'TreatyShare',
    'ReinsCurrency',
]

OED_REINS_SCOPE_FIELDS = [
    'ReinsNumber',
    'PortNumber',
    'AccNumber',
    'PolNumber',
    'LocGroup',
    'LocNumber',
    'CedantName',
    'ProducerName',
    'LOB',
    'CountryCode',
    'ReinsTag',
    'CededPercent'
]



import logging
import os
import shutil
import subprocess

from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from .bash import (bash_wrapper, create_bash_analysis,
                   create_bash_outputs, genbash)


@oasis_log()
def run(analysis_settings,
        number_of_processes=-1,
        set_alloc_rule_gul=None,
        set_alloc_rule_il=None,
        set_alloc_rule_ri=None,
        run_debug=False,
        custom_gulcalc_cmd=None,
        custom_gulcalc_log_start=None,
        custom_gulcalc_log_finish=None,
        custom_get_getmodel_cmd=None,
        filename='run_ktools.sh',
        gul_legacy_stream=False,
        df_engine='oasis_data_manager.df_reader.reader.OasisPandasReader',
        model_df_engine=None,
        dynamic_footprint=False,
        **kwargs
        ):
    model_df_engine = model_df_engine or df_engine

    #  MOVED into bash_params #########################################
    #  keep here for the moment and refactor after testing
    #
    #  Example:
    #  from .bash import get_complex_model_cmd
    #  <var> = get_complex_model_cmd(custom_gulcalc_cmd, analysis_settings)
    #
    # If `given_gulcalc_cmd` is set then always run as a complex model
    # and raise an exception when not found in PATH
    if custom_gulcalc_cmd:
        if not shutil.which(custom_gulcalc_cmd):
            raise OasisException(
                'Run error: Custom Gulcalc command "{}" explicitly set but not found in path.'.format(custom_gulcalc_cmd)
            )
    # when not set then fallback to previous behaviour:
    # Check if a custom binary `<supplier>_<model>_gulcalc` exists in PATH
    else:
        inferred_gulcalc_cmd = "{}_{}_gulcalc".format(
            analysis_settings.get('model_supplier_id'),
            analysis_settings.get('model_name_id'))
        if shutil.which(inferred_gulcalc_cmd):
            custom_gulcalc_cmd = inferred_gulcalc_cmd

    # TODO: should be integrated into bash.py
    if custom_gulcalc_cmd:
        if not custom_get_getmodel_cmd:
            def custom_get_getmodel_cmd(
                number_of_samples,
                gul_threshold,
                gul_legacy_stream,
                use_random_number_file,
                coverage_output,
                item_output,
                process_id,
                max_process_id,
                gul_alloc_rule,
                stderr_guard,
                **kwargs
            ):

                cmd = "{} -e {} {} -a {} -p {}".format(
                    custom_gulcalc_cmd,
                    process_id,
                    max_process_id,
                    os.path.abspath("analysis_settings.json"),
                    "input")
                if gul_legacy_stream and coverage_output != '':
                    cmd = '{} -c {}'.format(cmd, coverage_output)
                if item_output != '':
                    cmd = '{} -i {}'.format(cmd, item_output)
                if stderr_guard:
                    cmd = '({}) 2>> log/gul_stderror.err'.format(cmd)

                return cmd
        else:
            custom_get_getmodel_cmd = None

    ###########################################################

    # Calls run_analysis + run_outputs in a single script
    genbash(
        number_of_processes,
        analysis_settings,
        gul_alloc_rule=set_alloc_rule_gul,
        il_alloc_rule=set_alloc_rule_il,
        ri_alloc_rule=set_alloc_rule_ri,
        gul_legacy_stream=gul_legacy_stream,
        bash_trace=run_debug,
        filename=filename,
        _get_getmodel_cmd=custom_get_getmodel_cmd,
        custom_gulcalc_log_start=custom_gulcalc_log_start,
        custom_gulcalc_log_finish=custom_gulcalc_log_finish,
        model_df_engine=model_df_engine,
        dynamic_footprint=dynamic_footprint,
        **kwargs,
    )
    bash_trace = subprocess.check_output(['bash', filename])
    logging.info(bash_trace.decode('utf-8'))


@oasis_log()
def run_analysis(**params):
    with bash_wrapper(params['filename'],
                      params['bash_trace'],
                      params['stderr_guard'],
                      log_sub_dir=params.get("process_number", None),
                      process_number=params.get("process_number", None)):
        create_bash_analysis(**params)

    bash_trace = subprocess.check_output(['bash', params['filename']]).decode('utf-8')
    logging.info(bash_trace)
    return params['fifo_queue_dir'], bash_trace


@oasis_log()
def run_outputs(**params):
    with bash_wrapper(params['filename'], params['bash_trace'], params['stderr_guard'], log_sub_dir='out'):
        create_bash_outputs(**params)
    bash_trace = subprocess.check_output(['bash', params['filename']]).decode('utf-8')
    logging.info(bash_trace)
    return bash_trace



__all__ = [
    'PERILS',
    'PERIL_GROUPS',
]

from collections import OrderedDict
import pandas as pd

PRL_BBF = 'BBF'
PRL_BFR = 'BFR'
PRL_BSK = 'BSK'
PRL_MNT = 'MNT'
PRL_MTR = 'MTR'
PRL_ORF = 'ORF'
PRL_OSF = 'OSF'
PRL_QEQ = 'QEQ'
PRL_QFF = 'QFF'
PRL_QLF = 'QLF'
PRL_QLS = 'QLS'
PRL_QSL = 'QSL'
PRL_QTS = 'QTS'
PRL_WEC = 'WEC'
PRL_WSS = 'WSS'
PRL_WTC = 'WTC'
PRL_XHL = 'XHL'
PRL_XLT = 'XLT'
PRL_XSL = 'XSL'
PRL_XTD = 'XTD'
PRL_ZFZ = 'ZFZ'
PRL_ZIC = 'ZIC'
PRL_ZSN = 'ZSN'
PRL_ZST = 'ZST'

PRL_GRP_AA1 = 'AA1'
PRL_GRP_BB1 = 'BB1'
PRL_GRP_MM1 = 'MM1'
PRL_GRP_OO1 = 'OO1'
PRL_GRP_QQ1 = 'QQ1'
PRL_GRP_WW1 = 'WW1'
PRL_GRP_WW2 = 'WW2'
PRL_GRP_XX1 = 'XX1'
PRL_GRP_XZ1 = 'XZ1'
PRL_GRP_ZZ1 = 'ZZ1'

PERIL_GROUPS = OrderedDict({
    'all': {'id': PRL_GRP_AA1, 'desc': 'All perils', 'peril_ids': [PRL_QEQ, PRL_QFF, PRL_QTS, PRL_QSL, PRL_QLS, PRL_QLF, PRL_WTC, PRL_WEC, PRL_WSS, PRL_ORF, PRL_OSF, PRL_XSL, PRL_XTD, PRL_XHL, PRL_ZSN, PRL_ZIC, PRL_ZFZ, PRL_BFR, PRL_BBF, PRL_MNT, PRL_MTR, PRL_XLT, PRL_ZST, PRL_BSK]},
    'wildfire w/ smoke': {'id': PRL_GRP_BB1, 'desc': 'Wildfire with smoke', 'peril_ids': [PRL_BBF, PRL_BSK]},
    'terrorism': {'id': PRL_GRP_MM1, 'desc': 'Terrorism', 'peril_ids': [PRL_MNT, PRL_MTR]},
    'flood w/o storm surge': {'id': PRL_GRP_OO1, 'desc': 'Flood w/o storm surge', 'peril_ids': [PRL_ORF, PRL_OSF]},
    'earthquake': {'id': PRL_GRP_QQ1, 'desc': 'All EQ perils', 'peril_ids': [PRL_QEQ, PRL_QFF, PRL_QTS, PRL_QSL, PRL_QLS, PRL_QLF]},
    'windstorm w/ storm surge': {'id': PRL_GRP_WW1, 'desc': 'Windstorm with storm surge', 'peril_ids': [PRL_WTC, PRL_WEC, PRL_WSS]},
    'windstorm w/o storm surge': {'id': PRL_GRP_WW2, 'desc': 'Windstorm w/o storm surge', 'peril_ids': [PRL_WTC, PRL_WEC]},
    'convective storm': {'id': PRL_GRP_XX1, 'desc': 'Convective Storm', 'peril_ids': [PRL_XSL, PRL_XTD, PRL_XHL, PRL_XLT]},
    'convective storm rms': {'id': PRL_GRP_XZ1, 'desc': 'Convective storm (incl winter storm) - for RMS users', 'peril_ids': [PRL_XSL, PRL_XTD, PRL_XHL, PRL_ZSN, PRL_ZIC, PRL_ZFZ, PRL_XLT, PRL_ZST]},
    'winter storm': {'id': PRL_GRP_ZZ1, 'desc': 'Winter storm', 'peril_ids': [PRL_ZSN, PRL_ZIC, PRL_ZFZ, PRL_ZST]}
})

PERILS = OrderedDict({
    'wildfire': {'id': PRL_BBF, 'desc': 'Wildfire / Bushfire', 'group_peril': 'wildfire w/ smoke'},
    'noncat': {'id': PRL_BFR, 'desc': 'NonCat', 'group_peril': 'wildfire w/ smoke'},
    'smoke': {'id': PRL_BSK, 'desc': 'Smoke', 'group_peril': 'wildfire w/ smoke'},
    'nbcr terrorism': {'id': PRL_MNT, 'desc': 'NBCR Terrorism', 'group_peril': 'terrorism'},
    'terrorism': {'id': PRL_MTR, 'desc': 'Conventional Terrorism', 'group_peril': 'terrorism'},
    'river flood': {'id': PRL_ORF, 'desc': 'River / Fluvial Flood', 'group_peril': 'flood w/o storm surge'},
    'flash flood': {'id': PRL_OSF, 'desc': 'Flash / Surface / Pluvial Flood', 'group_peril': 'flood w/o storm surge'},
    'earthquake': {'id': PRL_QEQ, 'desc': 'Earthquake - Shake only', 'group_peril': 'earthquake'},
    'fire following': {'id': PRL_QFF, 'desc': 'Fire Following', 'group_peril': 'earthquake'},
    'liquefaction': {'id': PRL_QLF, 'desc': 'Liquefaction', 'group_peril': 'earthquake'},
    'landslide': {'id': PRL_QLS, 'desc': 'Landslide', 'group_peril': 'earthquake'},
    'sprinkler leakage': {'id': PRL_QSL, 'desc': 'Sprinkler Leakage', 'group_peril': 'earthquake'},
    'tsunami': {'id': PRL_QTS, 'desc': 'Tsunami', 'group_peril': 'earthquake'},
    'extra tropical cyclone': {'id': PRL_WEC, 'desc': 'Extra Tropical Cyclone', 'group_peril': 'windstorm w/o storm surge'},
    'storm surge': {'id': PRL_WSS, 'desc': 'Storm Surge', 'group_peril': 'windstorm with storm surge'},
    'tropical cyclone': {'id': PRL_WTC, 'desc': 'Tropical Cyclone', 'group_peril': 'windstorm w/o storm surge'},
    'hail': {'id': PRL_XHL, 'desc': 'Hail', 'group_peril': 'convective storm rms'},
    'lightning': {'id': PRL_XLT, 'desc': 'Lightning', 'group_peril': 'convective storm'},
    'convective wind': {'id': PRL_XSL, 'desc': 'Straight-line / other convective wind', 'group_peril': 'convective storm'},
    'tornado': {'id': PRL_XTD, 'desc': 'Tornado', 'group_peril': 'convective storm'},
    'freeze': {'id': PRL_ZFZ, 'desc': 'Freeze', 'group_peril': 'winter storm'},
    'ice': {'id': PRL_ZIC, 'desc': 'Ice', 'group_peril': 'winter storm'},
    'snow': {'id': PRL_ZSN, 'desc': 'Snow', 'eqv_oasis_peril': None, 'group_peril': 'winter storm'},
    'winterstorm wind': {'id': PRL_ZST, 'desc': 'Winterstorm Wind', 'group_peril': 'winter storm'}
})


def get_peril_groups_df():
    res = []
    for peril in PERILS.values():
        res.append((peril['id'], peril['id']))

    for peril_group in PERIL_GROUPS.values():
        for peril in peril_group['peril_ids']:
            res.append((peril_group['id'], peril))

    return pd.DataFrame(res, columns=['peril_group_id', 'peril_id'])



#!/usr/bin/env python

from . import manager, logger

import argparse
import logging

parser = argparse.ArgumentParser()
parser.add_argument(
    '-f', '--secondary-factor',
    help='optional relative secondary factor within range [0, 1]', default=1.0,
    type=float
)
parser.add_argument(
    '-F', '--uniform-factor',
    help='optional uniform post loss amplification factor', default=0.0,
    type=float
)
parser.add_argument('-i', '--file-in', help='name of the input file')
parser.add_argument('-o', '--file-out', help='name of the output file')
parser.add_argument(
    '-P', '--input-path', help='path to amplifications.bin',
    default='input'
)
parser.add_argument(
    '-p', '--static-path', help='path to lossfactors.bin', default='static'
)
parser.add_argument(
    '-r', '--run-dir', help='path to the run directory', default='.'
)
parser.add_argument(
    '-v', '--logging-level',
    help='logging level (debug:10, info:20, warning:30, error:40, critical:50)',
    default=30, type=int
)


def main():
    kwargs = vars(parser.parse_args())

    # Add handler to pla logger
    ch = logging.StreamHandler()
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
    )
    ch.setFormatter(formatter)
    logger.addHandler(ch)
    logging_level = kwargs.pop('logging_level')
    logger.setLevel(logging_level)

    if kwargs['secondary_factor'] < 0.0 or kwargs['secondary_factor'] > 1.0:
        logger.error(f'Secondary factor {kwargs["secondary_factor"]} must lie within range [0, 1]')
        SystemExit(1)

    if kwargs['uniform_factor'] < 0.0:
        logger.error(f'Uniform factor {kwargs["uniform_factor"]} must be positive value')
        SystemExit(1)

    if kwargs['secondary_factor'] < 1.0 and kwargs['uniform_factor'] > 0.0:
        logger.warning('Relative secondary and uniform factors are incompatible')
        logger.info('Ignoring relative secondary factor')
        kwargs['secondary_factor'] = 1.0

    manager.run(**kwargs)


if __name__ == '__main__':
    main()



import numpy as np

BUFFER_SIZE = 65536
DATA_SIZE = 8   # 4 bytes * 2 values
N_PAIRS = BUFFER_SIZE // DATA_SIZE
FILE_HEADER_SIZE = 4   # 4-byte integer

# Numpy data types for reading binary files
event_count_dtype = np.dtype([('event_id', 'i4'), ('count', 'i4')])
amp_factor_dtype = np.dtype([('amplification_id', 'i4'), ('factor', 'f4')])

# File names for creating dictionary of factors
AMPLIFICATIONS_FILE_NAME = 'amplifications.bin'
LOSS_FACTORS_FILE_NAME = 'lossfactors.bin'




__all__ = [
    'GenerateOasisLosses'
]

from tqdm import tqdm

from ..base import ComputationStep
from ..generate.losses import GenerateLosses
from ..hooks.pre_loss import PreLoss
from ..hooks.post_analysis import PostAnalysis


class GenerateOasisLosses(ComputationStep):
    """
    Run Oasis file geneartion with optional PreAnalysis hook.
    """

    # Override params
    step_params = [
        {'name': 'pre_loss_module', 'required': False, 'is_path': True, 'pre_exist': True,
         'help': 'Pre-Loss module path'},
        {'name': 'post_analysis_module', 'required': False, 'is_path': True, 'pre_exist': True,
         'help': 'Post-Analysis module path'},
    ]
    # Add params from each sub command not in 'step_params'
    chained_commands = [
        PreLoss,
        GenerateLosses,
        PostAnalysis,
    ]

    def run(self):

        # setup output dir
        if not self.model_run_dir:
            self.model_run_dir = GenerateLosses._get_output_dir(self)
        self.kwargs['model_run_dir'] = self.model_run_dir

        # Run chain
        if self.pre_loss_module:
            cmds = [(PreLoss, self.kwargs)]
        else:
            cmds = []
        cmds += [(GenerateLosses, self.kwargs)]
        if self.post_analysis_module:
            cmds += [(PostAnalysis, self.kwargs)]

        with tqdm(total=len(cmds)) as pbar:
            for cmd in cmds:
                cmd[0](**cmd[1]).run()
                pbar.update(1)

        self.logger.info(f'Losses generated in {self.model_run_dir}')



__version__ = '2.3.15'

import sys
from importlib.abc import MetaPathFinder, Loader
from importlib.util import spec_from_loader
import importlib
import warnings

import logging
from logging import NullHandler

logger = logging.getLogger(__name__)
handler = NullHandler()
handler.name = 'oasislmf'
logger.addHandler(handler)


class MyLoader(Loader):

    def __init__(self, sub_module):
        self.sub_module = sub_module

    def module_repr(self, module):
        return repr(module)

    def exec_module(self, module):
        old_name = module.__name__
        new_name = f"oasislmf.{self.sub_module}"
        sys.modules[old_name] = importlib.import_module(new_name)
        return sys.modules[old_name]


class MyImport(MetaPathFinder):
    """ Support alias of depreciated sub-modules

        * model_execution   -> execution
        * model_preparation -> preparation
        * api               -> platform

        Example:
            `from oasislmf.model_execution.bash import genbash`
                is the same as calling the new name
            `from oasislmf.execution.bash import genbash`
            https://docs.python.org/3/library/importlib.html#importlib.machinery.PathFinder
    """

    def __init__(self):
        self.depricated_modules = {
            "model_execution": "execution",
            "model_preparation": "preparation",
            "api": "platform",
            "platform": "platform_api"
        }

    def find_spec(self, fullname, path=None, target=None):
        import_path = fullname.split(".", 1)
        if fullname.startswith("oasislmf") and len(import_path) > 1:
            import_path = import_path[1]
            for deprecated in self.depricated_modules:
                if deprecated == import_path or import_path.startswith(deprecated + '.'):
                    with warnings.catch_warnings():
                        warnings.simplefilter("always")
                        warnings.warn(
                            f"imports from 'oasislmf.{deprecated}' are deprecated. Import by using 'oasislmf.{self.depricated_modules[deprecated]}' instead."
                        )
                    import_path = import_path.replace(deprecated, self.depricated_modules[deprecated])

            return spec_from_loader(fullname, MyLoader(import_path))


sys.meta_path.append(MyImport())



import logging
import os
import shutil

import numpy as np
import pandas as pd

from ods_tools.oed import fill_empty

from ..utils.exceptions import OasisException
from ..utils.log import oasis_log
from ..utils.data import get_dataframe
from . import oed


REINS_RISK_LEVEL_XREF_COLUMN_MAP = {
    oed.REINS_RISK_LEVEL_LOCATION_GROUP: ["LocGroup", "PortNumber", "AccNumber", "PolNumber", "LocNumber"],
    oed.REINS_RISK_LEVEL_LOCATION: ["PortNumber", "AccNumber", "LocNumber", "PolNumber"]
}
XREF_COLUMN_DEFAULT = ["PortNumber", "AccNumber", "PolNumber", "LocNumber"]

RISK_LEVEL_FIELD_MAP = {
    oed.REINS_RISK_LEVEL_PORTFOLIO: ['PortNumber'],
    oed.REINS_RISK_LEVEL_ACCOUNT: ['PortNumber', 'AccNumber'],
    oed.REINS_RISK_LEVEL_POLICY: ['PortNumber', 'AccNumber', 'PolNumber'],
    oed.REINS_RISK_LEVEL_LOCATION_GROUP: ['LocGroup'],
    oed.REINS_RISK_LEVEL_LOCATION: ['PortNumber', 'AccNumber', 'LocNumber']
}
RISK_LEVEL_ALL_FIELDS = ['PortNumber', 'AccNumber', 'PolNumber', 'LocGroup', 'LocNumber']
FILTER_LEVEL_EXTRA_FIELDS = ['CedantName', 'ProducerName', 'LOB', 'CountryCode', 'ReinsTag']

CHECK_RI_SCOPE_MAP = {
    oed.REINS_RISK_LEVEL_ACCOUNT: {'in': ['AccNumber'], 'out': ['PolNumber', 'LocNumber']},
    oed.REINS_RISK_LEVEL_POLICY: {'in': ['AccNumber', 'PolNumber'], 'out': ['LocNumber']},
    oed.REINS_RISK_LEVEL_LOCATION: {'in': ['AccNumber', 'LocNumber']},
    oed.REINS_RISK_LEVEL_LOCATION_GROUP: {'in': ['LocGroup']},
}

REINS_TYPE_EXACT_MATCH = [oed.REINS_TYPE_FAC, oed.REINS_TYPE_SURPLUS_SHARE]

NO_LOSS_PROFILE_ID = 1
PASSTHROUGH_PROFILE_ID = 2

NO_LOSS_PROFILE = dict(
    calcrule_id=oed.CALCRULE_ID_LIMIT_ONLY,
    deductible1=0.0,  # Not used
    deductible2=0.0,  # Not used
    deductible3=0.0,  # Not used
    attachment=0.0,   # Not used
    limit=0.0,
    share1=0.0,       # Not used
    share2=0.0,       # Not used
    share3=0.0        # Not used
)

PASSTHROUGH_PROFILE = dict(
    calcrule_id=oed.CALCRULE_ID_DEDUCTIBLE_ONLY,
    deductible1=0.0,
    deductible2=0.0,  # Not used
    deductible3=0.0,  # Not used
    attachment=0.0,   # Not used
    limit=0.0,        # Not used
    share1=0.0,       # Not used
    share2=0.0,       # Not used
    share3=0.0        # Not used
)


ITEM_LEVEL_ID = 1
FILTER_LEVEL_ID = 2
RISK_LEVEL_ID = 3
PROGRAM_LEVEL_ID = 4

FM_TERMS_PER_REINS_TYPE = {
    oed.REINS_TYPE_FAC: {
        RISK_LEVEL_ID: {
            'calcrule_id': {'default': oed.CALCRULE_ID_OCCURRENCE_CATASTROPHE_EXCESS_OF_LOSS},
            'deductible1': {'default': 0.},
            'deductible2': {'default': 0.},
            'deductible3': {'default': 0.},
            'attachment': {'oed_col': 'RiskAttachment', 'default': 0.},
            'limit': {'oed_col': 'RiskLimit', 'default': oed.LARGE_VALUE, 'to_default': [0.]},
            'share1': {'oed_col': 'CededPercent', 'default': 1.},
            'share2': {'oed_col': 'PlacedPercent', 'default': 1.},
            'share3': {'default': 1.},
        }
    },
    oed.REINS_TYPE_PER_RISK: {
        RISK_LEVEL_ID: {
            'calcrule_id': {'default': oed.CALCRULE_ID_OCCURRENCE_CATASTROPHE_EXCESS_OF_LOSS},
            'deductible1': {'default': 0.},
            'deductible2': {'default': 0.},
            'deductible3': {'default': 0.},
            'attachment': {'oed_col': 'RiskAttachment', 'default': 0.},
            'limit': {'oed_col': 'RiskLimit', 'default': oed.LARGE_VALUE, 'to_default': [0.]},
            'share1': {'oed_col': 'CededPercent', 'default': 1.},
            'share2': {'default': 1.},
            'share3': {'default': 1.},
        },
        PROGRAM_LEVEL_ID: {
            'calcrule_id': {'default': oed.CALCRULE_ID_OCCURRENCE_CATASTROPHE_EXCESS_OF_LOSS},
            'deductible1': {'default': 0.},
            'deductible2': {'default': 0.},
            'deductible3': {'default': 0.},
            'attachment': {'oed_col': 'OccAttachment', 'default': 0.},
            'limit': {'oed_col': 'OccLimit', 'default': oed.LARGE_VALUE, 'to_default': [0.]},
            'share1': {'default': 1.},
            'share2': {'oed_col': 'PlacedPercent', 'default': 1.},
            'share3': {'default': 1.},
        }
    },
    oed.REINS_TYPE_QUOTA_SHARE: {
        RISK_LEVEL_ID: {
            'calcrule_id': {'default': oed.CALCRULE_ID_OCCURRENCE_CATASTROPHE_EXCESS_OF_LOSS},
            'deductible1': {'default': 0.},
            'deductible2': {'default': 0.},
            'deductible3': {'default': 0.},
            'attachment': {'default': 0.},
            'limit': {'oed_col': 'RiskLimit', 'default': oed.LARGE_VALUE, 'to_default': [0.]},
            'share1': {'oed_col': 'CededPercent', 'default': 1.},
            'share2': {'default': 1.},
            'share3': {'default': 1.},
        },
        PROGRAM_LEVEL_ID: {
            'calcrule_id': {'default': oed.CALCRULE_ID_OCCURRENCE_LIMIT_AND_SHARE},
            'deductible1': {'default': 0.},
            'deductible2': {'default': 0.},
            'deductible3': {'default': 0.},
            'attachment': {'default': 0.},
            'limit': {'oed_col': 'OccLimit', 'default': oed.LARGE_VALUE, 'to_default': [0.]},
            'share1': {'default': 0.},
            'share2': {'oed_col': 'PlacedPercent', 'default': 1.},
            'share3': {'default': 1.},
        }
    },
    oed.REINS_TYPE_SURPLUS_SHARE: {
        RISK_LEVEL_ID: {
            'calcrule_id': {'default': oed.CALCRULE_ID_OCCURRENCE_CATASTROPHE_EXCESS_OF_LOSS},
            'deductible1': {'default': 0.},
            'deductible2': {'default': 0.},
            'deductible3': {'default': 0.},
            'attachment': {'oed_col': 'RiskAttachment', 'default': 0.},
            'limit': {'oed_col': 'RiskLimit', 'default': oed.LARGE_VALUE, 'to_default': [0.]},
            'share1': {'oed_col': 'CededPercent_scope', 'default': 1.},
            'share2': {'default': 1.},
            'share3': {'default': 1.},
        },
        PROGRAM_LEVEL_ID: {
            'calcrule_id': {'default': oed.CALCRULE_ID_OCCURRENCE_LIMIT_AND_SHARE},
            'deductible1': {'default': 0.},
            'deductible2': {'default': 0.},
            'deductible3': {'default': 0.},
            'attachment': {'default': 0.},
            'limit': {'oed_col': 'OccLimit', 'default': oed.LARGE_VALUE, 'to_default': [0.]},
            'share1': {'default': 0.},
            'share2': {'oed_col': 'PlacedPercent', 'default': 1.},
            'share3': {'default': 1.},
        }
    },
    oed.REINS_TYPE_CAT_XL: {
        PROGRAM_LEVEL_ID: {
            'calcrule_id': {'default': oed.CALCRULE_ID_OCCURRENCE_CATASTROPHE_EXCESS_OF_LOSS},
            'deductible1': {'default': 0.},
            'deductible2': {'default': 0.},
            'deductible3': {'default': 0.},
            'attachment': {'oed_col': 'OccAttachment', 'default': 0.},
            'limit': {'oed_col': 'OccLimit', 'default': oed.LARGE_VALUE, 'to_default': [0.]},
            'share1': {'oed_col': 'CededPercent', 'default': 1.},
            'share2': {'oed_col': 'PlacedPercent', 'default': 1.},
            'share3': {'default': 1.},
        }
    }
}


def create_risk_level_profile_id(ri_df, profile_map_df, fm_profile_df, reins_type, risk_level, fm_level_id, logger):
    """
    Create new profile id from reinsurance in ri_df corresponding to reins_type.
    Add them to fm_profile_df and match the profile_ids in ri_df and profile_map_df
    Args:
        ri_df: ri info and scope
        profile_map_df: tree structure df representing each ri fm levels
        fm_profile_df: df containing all the profiles
        reins_type: type of reinsurance (one of oed.REINS_TYPES)
        risk_level: level of the reinsurance terms (one of oed.REINS_RISK_LEVELS)
        fm_level_id: fm level in profile_map_df

    Returns:
        fm_profile_df: updated version of fm_profile_df
    """
    reins_type_filter = ri_df['ReinsType'] == reins_type
    if not reins_type_filter.any():
        return fm_profile_df

    # create new fm profile from ri terms corresponding to the reins_type
    ri_term_map = {term_info['oed_col']: term for term, term_info in FM_TERMS_PER_REINS_TYPE[reins_type].get(fm_level_id, {}).items()
                   if 'oed_col' in term_info}

    if ri_term_map:
        # create a profile_id for each unique term combination
        ri_df.loc[reins_type_filter, 'profile_id'] = pd.factorize(pd._libs.lib.fast_zip([
            ri_df[reins_type_filter][col].values for col in ri_term_map
        ]))[0] + 1 + fm_profile_df['profile_id'].max()

        # create complete profile from terms and add it to fm_profile_df
        cur_fm_profiles = ri_df[reins_type_filter][['profile_id'] + list(ri_term_map)].rename(columns=ri_term_map)
        for term, term_info in FM_TERMS_PER_REINS_TYPE[reins_type][fm_level_id].items():
            if term not in cur_fm_profiles:
                cur_fm_profiles[term] = term_info['default']
            else:
                fill_empty(cur_fm_profiles, [term], term_info['default'])
                cur_fm_profiles.loc[cur_fm_profiles[term].isin(term_info.get('to_default', [])), term] = term_info['default']
        cur_fm_profiles.drop_duplicates(inplace=True)
        if not cur_fm_profiles.empty:
            fm_profile_df = pd.concat([fm_profile_df, cur_fm_profiles], ignore_index=True)

    else:  # No terms at risk level
        ri_df.loc[reins_type_filter, 'profile_id'] = PASSTHROUGH_PROFILE_ID

    # update profile_map profile_id for filter and risk level
    layer_id_set = set(ri_df[reins_type_filter]['layer_id'])
    these_profile_map_layers = profile_map_df[profile_map_df['layer_id'].isin(layer_id_set)]

    # filter level
    if fm_level_id == RISK_LEVEL_ID:
        ri_filter_fields = RISK_LEVEL_ALL_FIELDS + [field for field in FILTER_LEVEL_EXTRA_FIELDS if field in ri_df]
        ri_filter_valid_fields = [field + '_valid' for field in ri_filter_fields]
        ri_date_fields = ['ReinsInceptionDate', 'ReinsExpiryDate', 'AttachmentBasis']
        merge_on = ['layer_id']
        if reins_type in REINS_TYPE_EXACT_MATCH:
            merge_on += RISK_LEVEL_FIELD_MAP[risk_level]
        filter_df = (
            these_profile_map_layers[these_profile_map_layers['level_id'] == FILTER_LEVEL_ID]
            .reset_index()
            .merge(
                ri_df[reins_type_filter][ri_filter_fields + ri_filter_valid_fields + ['layer_id'] + ri_date_fields],
                how='inner',
                on=merge_on
            )
        )

        def _match(row):
            for field in ri_filter_fields:
                if (field not in merge_on
                        and row[f'{field}_valid'] and row[f'{field}_x'] != row[f'{field}_y']):
                    return False

            # Risk Attaching filter for reinsurance
            if "AttachmentBasis" in row and row["AttachmentBasis"] == "RA":
                if row["ReinsInceptionDate"] == "" or row["ReinsExpiryDate"] == "":
                    error_msg = "Error: ReinsInceptionDate/ReinsExpiryDate missing, cannot use AttachmentBasis [RA]. Please check the ri_info file"
                    raise OasisException(error_msg)
                elif row["PolInceptionDate"] == "":
                    acc_info = {
                        field: row[f'{field}_x'] if f'{field}_x' in row else row[f'{field}']
                        for field in RISK_LEVEL_FIELD_MAP[oed.REINS_RISK_LEVEL_ACCOUNT]
                        if f'{field}_x' in row or f'{field}' in row
                    }
                    error_msg = f"Error: PolInceptionDate missing for {acc_info}, cannot use AttachmentBasis [RA]. Please check the account file"
                    raise OasisException(error_msg)
                else:
                    if row["PolInceptionDate"] < row["ReinsInceptionDate"] or row["ReinsExpiryDate"] < row["PolInceptionDate"]:
                        return False

            return True
        profile_map_df.loc[np.unique(filter_df.loc[filter_df.apply(_match, axis=1), 'index']), 'profile_id'] = PASSTHROUGH_PROFILE_ID

    # Risk level
    layer_filter = reins_type_filter
    match_fields = list(RISK_LEVEL_FIELD_MAP[risk_level]) if fm_level_id == RISK_LEVEL_ID else []

    while layer_filter.any():
        ri_filter = layer_filter & (ri_df[match_fields] != '').all(axis=1)

        df_ = these_profile_map_layers[these_profile_map_layers['level_id'] == fm_level_id].reset_index().merge(
            ri_df[ri_filter][match_fields + ['layer_id', 'profile_id']].drop_duplicates(),
            how='inner', on=match_fields + ['layer_id'], suffixes=['', '_y']
        ).set_index('index')

        profile_map_df.loc[df_.index, 'profile_id'] = df_['profile_id_y']

        # match field are ordered by granularity, if an exact match is not needed we pop the most granular column and look for more general match
        if match_fields:
            if reins_type in REINS_TYPE_EXACT_MATCH:
                break
            else:
                layer_filter = layer_filter & (ri_df[match_fields.pop()] == '')
        else:
            break

    return fm_profile_df


def check_ri_scope_filter(ri_df, risk_level):
    # For some treaty types the scope filter must match exactly
    column_in = CHECK_RI_SCOPE_MAP.get(risk_level, {}).get('in', [])
    column_out = CHECK_RI_SCOPE_MAP.get(risk_level, {}).get('out', [])
    return (
        (~ri_df['ReinsType'].isin(REINS_TYPE_EXACT_MATCH))
        | (
            (column_in == [] or (ri_df[column_in] != "").all(axis=1))
            & (column_out == [] or (ri_df[column_out] == "").all(axis=1))
        )
    )


def get_xref_df(xref_descriptions_df, risk_level):
    """
    Build the cross-reference dataframe, which serves as a representation
    of the insurance programme depending on the reinsurance risk level.
    Dataframes for programme, risk, filter and items levels are created.
    The fields agg_id, level_id and to_agg_id (agg_id_to), which are used
    to construct the FM Programmes structure, are assigned. The
    aforementioned dataframes are concatenated to form a single dataframe
    called xref_df, which is returned. The returned dataframe features the
    fields necessary for the assignment of profile IDs.
    Args:
        xref_descriptions_df: Fm summary mapping enhanced by relevant information from Loc and Account
        risk_level: risk_level

    Returns:
        df_levels: list of dataframes, one per fm level
    """

    xref_descriptions = xref_descriptions_df.sort_values(by=REINS_RISK_LEVEL_XREF_COLUMN_MAP.get(risk_level, XREF_COLUMN_DEFAULT))
    risk_level_fields = RISK_LEVEL_FIELD_MAP[risk_level]

    df_levels = dict()
    # Programme level
    df_levels['programme_level'] = pd.DataFrame(
        {'agg_id': 1, 'level_id': PROGRAM_LEVEL_ID, 'agg_id_to': 0},
        index=[0]
    )

    # Risk level
    risk_level_df = xref_descriptions.drop_duplicates(
        subset=risk_level_fields, keep='first'
    ).reset_index(drop=True)
    risk_level_df['agg_id'] = risk_level_df.index + 1
    risk_level_df['level_id'] = RISK_LEVEL_ID
    risk_level_df['agg_id_to'] = 1
    df_levels['risk_level'] = risk_level_df

    # Filter level
    filter_level_df = xref_descriptions.drop_duplicates(
        subset=RISK_LEVEL_ALL_FIELDS
    ).reset_index(drop=True)
    filter_level_df['agg_id'] = filter_level_df.index + 1
    filter_level_df['level_id'] = FILTER_LEVEL_ID
    filter_level_df = filter_level_df.merge(
        risk_level_df[risk_level_fields + ['agg_id']], how='left', on=risk_level_fields,
        suffixes=['', '_to']
    )
    df_levels['filter_level'] = filter_level_df

    # Item level
    item_level_df = xref_descriptions.reset_index(drop=True)
    item_level_df['agg_id'] = item_level_df['output_id']
    item_level_df['level_id'] = ITEM_LEVEL_ID
    item_level_df = item_level_df.merge(
        filter_level_df[RISK_LEVEL_ALL_FIELDS + ['agg_id']], how='left',
        on=RISK_LEVEL_ALL_FIELDS, suffixes=['', '_to']
    )
    df_levels['items_level'] = item_level_df

    return df_levels


def _log_dataframe(logger, df_dict, ri_name):
    if logger.level >= logging.DEBUG:
        for df_name, df_ in df_dict.items():
            logger.debug(f'{df_name}: {ri_name}:')
            logger.debug(df_)


@oasis_log
def write_files_for_reinsurance(ri_info_df, ri_scope_df, xref_descriptions_df, output_dir, fm_xref_fp, logger):
    """
    Create the Oasis structures - FM Programmes, FM Profiles and FM Policy
    TCs - that represent the reinsurance structure.

    The cross-reference dataframe, which serves as a representation of the
    insurance programme depending on the reinsurance risk level, is built.
    Except facultative contracts, each contract is a
    separate layer. Profile IDs for the risk and filter levels are created
    using the merged reinsurance scope and info dataframes. These profile
    IDs are assigned according to some combination of the fields
    PortNumber, AccNumber, PolNumber, LocGroup and LocNumber, dependent on
    reinsurance risk level. Individual programme level profile IDs are
    assigned for each row of the reinsurance info dataframe. Finally, the
    Oasis structure is written out.
    """
    fm_xref_df = get_dataframe(fm_xref_fp)
    fm_xref_df['agg_id'] = range(1, 1 + len(fm_xref_df))
    fill_empty(ri_scope_df, RISK_LEVEL_ALL_FIELDS, '')

    reinsurance_index = 1
    inuring_metadata = {}
    for inuring_priority in range(1, ri_info_df['InuringPriority'].max() + 1):
        for risk_level in oed.REINS_RISK_LEVELS:
            cur_ri_info_df = ri_info_df[(ri_info_df['InuringPriority'] == inuring_priority) & (ri_info_df['RiskLevel'] == risk_level)]
            if cur_ri_info_df.empty:
                continue

            output_name = f"ri_{inuring_priority}_{risk_level}"
            df_levels = get_xref_df(xref_descriptions_df, risk_level)

            _log_dataframe(logger, df_levels, output_name)

            no_loss_profile = {'profile_id': NO_LOSS_PROFILE_ID, **NO_LOSS_PROFILE}
            pass_through_profile = {'profile_id': PASSTHROUGH_PROFILE_ID, **PASSTHROUGH_PROFILE}

            fm_profile_df = pd.DataFrame([no_loss_profile, pass_through_profile])

            xref_df = pd.concat(df_levels.values(), ignore_index=True)

            # Assign default profile IDs
            xref_df['profile_id'] = NO_LOSS_PROFILE_ID
            xref_df['profile_id'] = xref_df['profile_id'].where(
                xref_df['level_id'].isin([FILTER_LEVEL_ID, RISK_LEVEL_ID]),
                PASSTHROUGH_PROFILE_ID
            )

            # Merge RI info and scope dataframes, and assign layers
            # Use as few layers as possible for FAC
            # Otherwise separate layers for each contract
            logger.debug(
                'Merging RI info and scope dataframes and assigning layers'
            )
            risk_level_fields = RISK_LEVEL_FIELD_MAP[risk_level]
            ri_df = cur_ri_info_df.merge(ri_scope_df, on='ReinsNumber', suffixes=['', '_scope'])
            valid_rows = check_ri_scope_filter(ri_df, risk_level)
            if not valid_rows.all():
                raise OasisException(
                    f'Invalid combination of Risk Level and Reinsurance Type. Please check scope file:\n{ri_df[~valid_rows]}'
                )

            ri_df['layer_id'] = 0
            ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_FAC, 'layer_id'] = (ri_df.loc[ri_df['ReinsType'] == oed.REINS_TYPE_FAC]
                                                                               .groupby(risk_level_fields, observed=True).cumcount() + 1)
            ri_info_no_fac = cur_ri_info_df[cur_ri_info_df['ReinsType'] != oed.REINS_TYPE_FAC].reset_index(drop=True)
            ri_info_no_fac['layer_id'] = ri_info_no_fac.index + 1 + ri_df['layer_id'].max()
            ri_df = ri_df.merge(ri_info_no_fac, how='left', on=ri_info_no_fac.columns.to_list()[:-1], suffixes=['', '_y'])
            ri_df['layer_id'] = ri_df['layer_id'].where(ri_df['layer_id_y'].isna(), ri_df['layer_id_y'])
            ri_df = ri_df.drop('layer_id_y', axis=1)

            for field in RISK_LEVEL_ALL_FIELDS:
                ri_df[field + '_valid'] = (ri_df[field] != '')
            for field in FILTER_LEVEL_EXTRA_FIELDS:
                if field in ri_df.columns:
                    fill_empty(ri_df, [field], '')
                    ri_df[field + '_valid'] = (ri_df[field] != '')

            del ri_info_no_fac

            profile_maps = [xref_df.copy() for i in range(ri_df['layer_id'].max())]
            for i, df in enumerate(profile_maps):
                df['layer_id'] = i + 1
            profile_map_df = pd.concat(profile_maps, ignore_index=True)

            logger.debug('Creating risk level and filter level profile IDs:')
            for fm_level_id in [RISK_LEVEL_ID, PROGRAM_LEVEL_ID]:
                for reins_type in FM_TERMS_PER_REINS_TYPE:
                    logger.debug(f'level_id {fm_level_id}, {reins_type} profiles...')
                    fm_profile_df = create_risk_level_profile_id(ri_df, profile_map_df, fm_profile_df, reins_type, risk_level, fm_level_id, logger)

            ri_df['profile_id'] = ri_df['profile_id'].astype('int64')
            profile_map_df['profile_id'] = profile_map_df['profile_id'].astype('int64')
            fm_profile_df['profile_id'] = fm_profile_df['profile_id'].astype('int64')

            _log_dataframe(logger, {'ri_df': ri_df, 'profile_map_df': profile_map_df, 'fm_profile_df': fm_profile_df}, output_name)

            # create fm df
            fm_programme_df = xref_df[xref_df['agg_id_to'] != 0][['agg_id', 'level_id', 'agg_id_to']].reset_index(drop=True)
            fm_programme_df.columns = ['from_agg_id', 'level_id', 'to_agg_id']
            fm_profile_df = fm_profile_df.sort_values(by='profile_id').reset_index(drop=True)

            fm_policytc_df = profile_map_df[profile_map_df['level_id'] > 1][
                ['layer_id', 'level_id', 'agg_id', 'profile_id']
            ].reset_index(drop=True)
            fm_policytc_df['level_id'] = fm_policytc_df['level_id'] - 1
            # Net losses across all layers is associated to the max layer ID.
            fm_xref_df['layer_id'] = fm_policytc_df['layer_id'].max()

            _log_dataframe(logger, {'fm_programme_df': fm_programme_df, 'fm_profile_df': fm_profile_df,
                                    'fm_policytc_df': fm_policytc_df}, output_name)

            # Write out Oasis structure
            ri_output_dir = os.path.join(output_dir, "RI_{}".format(reinsurance_index))
            if os.path.exists(ri_output_dir):
                shutil.rmtree(ri_output_dir)
            os.makedirs(ri_output_dir)

            fm_programme_df.to_csv(
                os.path.join(ri_output_dir, "fm_programme.csv"), index=False)
            fm_profile_df.to_csv(
                os.path.join(ri_output_dir, "fm_profile.csv"), index=False)
            fm_policytc_df.to_csv(
                os.path.join(ri_output_dir, "fm_policytc.csv"), index=False)
            fm_xref_df.to_csv(
                os.path.join(ri_output_dir, "fm_xref.csv"), index=False)

            inuring_metadata[reinsurance_index] = {
                'inuring_priority': inuring_priority,
                'risk_level': risk_level,
                'directory': ri_output_dir
            }

            reinsurance_index = reinsurance_index + 1

    return inuring_metadata



import numba as nb
import numpy as np

from oasislmf.pytools.common.data import areaperil_int, oasis_float

# gul stream type
# probably need to set this dynamically depending on the stream type
gul_header = np.int32(1 | 2 << 24).tobytes()

Item = nb.from_dtype(np.dtype([('item_id', np.int32),
                               ('coverage_id', np.int32),
                               ('areaperil_id', areaperil_int),
                               ('vulnerability_id', np.int32),
                               ('group_id', np.int32)
                               ]))

ItemAdjustment = nb.from_dtype(np.dtype([('item_id', np.int32),
                                         ('intensity_adjustment', np.int32),
                                         ('return_period', np.int32)
                                         ]))

items_data_type = nb.from_dtype(np.dtype([('item_id', np.int32),
                                          ('damagecdf_i', np.int32),
                                          ('rng_index', np.int32)
                                          ]))

items_MC_data_type = nb.from_dtype(np.dtype([('item_id', np.int32),
                                             ('item_idx', np.int32),
                                             ('hazcdf_i', np.int32),
                                             ('rng_index', np.int32),
                                             ('hazard_rng_index', np.int32),
                                             ('intensity_adjustment', np.int32),
                                             ('return_period', np.int32)
                                             ]))

VulnCdfLookup = nb.from_dtype(np.dtype([('start', np.int32),
                                        ('length', np.int32)]))


coverage_type = nb.from_dtype(np.dtype([('tiv', np.float64),
                                        ('max_items', np.int32),
                                        ('start_items', np.int32),
                                        ('cur_items', np.int32)
                                        ]))

NP_BASE_ARRAY_SIZE = 8


ITEM_MAP_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int32))
# ITEM_MAP_VALUE_TYPE = nb.types.UniTuple(nb.types.int32, 1)
ITEM_MAP_VALUE_TYPE = nb.types.int64
AREAPERIL_TO_EFF_VULN_KEY_TYPE = nb.types.Tuple((nb.from_dtype(areaperil_int), nb.types.int64))
AREAPERIL_TO_EFF_VULN_VALUE_TYPE = nb.types.UniTuple(nb.types.int32, 2)

# compute the relative size of oasis_float and areaperil_int vs int32
oasis_float_to_int32_size = oasis_float.itemsize // np.int32().itemsize
areaperil_int_to_int32_size = areaperil_int.itemsize // np.int32().itemsize

haz_cdf_type = nb.from_dtype(np.dtype([('probability', oasis_float),
                                       ('intensity_bin_id', np.int32)]))

ProbMean = nb.from_dtype(np.dtype([('prob_to', oasis_float),
                                   ('bin_mean', oasis_float)
                                   ]))
ProbMean_size = ProbMean.size

damagecdfrec_stream = nb.from_dtype(np.dtype([('event_id', np.int32),
                                              ('areaperil_id', areaperil_int),
                                              ('vulnerability_id', np.int32)
                                              ]))

damagecdfrec = nb.from_dtype(np.dtype([('areaperil_id', areaperil_int),
                                       ('vulnerability_id', np.int32)
                                       ]))


gulSampleslevelHeader = nb.from_dtype(np.dtype([('event_id', 'i4'),
                                                ('item_id', 'i4'),
                                                ]))
gulSampleslevelHeader_size = gulSampleslevelHeader.size

gulSampleslevelRec = nb.from_dtype(np.dtype([('sidx', 'i4'),
                                             ('loss', oasis_float),
                                             ]))
gulSampleslevelRec_size = gulSampleslevelRec.size


Keys = {'LocID': np.int32,
        'PerilID': 'category',
        'CoverageTypeID': np.int32,
        'AreaPerilID': areaperil_int,
        'VulnerabilityID': np.int32}


NormInversionParameters = nb.from_dtype(np.dtype([('x_min', np.float32),
                                                  ('x_max', np.float32),
                                                  ('N', np.int32),
                                                  ('cdf_min', np.float32),
                                                  ('cdf_max', np.float32),
                                                  ]))



import os
import numba as nb
import numpy as np
import pandas as pd

oasis_int = np.dtype(os.environ.get('OASIS_INT', 'i4'))
nb_oasis_int = nb.from_dtype(oasis_int)
oasis_int_size = oasis_int.itemsize

oasis_float = np.dtype(os.environ.get('OASIS_FLOAT', 'f4'))
nb_oasis_float = nb.from_dtype(oasis_float)
oasis_float_size = oasis_float.itemsize

areaperil_int = np.dtype(os.environ.get('AREAPERIL_TYPE', 'u4'))
nb_areaperil_int = nb.from_dtype(areaperil_int)
areaperil_int_size = areaperil_int.itemsize

null_index = oasis_int.type(-1)


summary_xref_dtype = np.dtype([('item_id', 'i4'), ('summary_id', 'i4'), ('summary_set_id', 'i4')])

# financial structure static input dtypes
fm_programme_dtype = np.dtype([('from_agg_id', 'i4'), ('level_id', 'i4'), ('to_agg_id', 'i4')])
fm_policytc_dtype = np.dtype([('level_id', 'i4'), ('agg_id', 'i4'), ('layer_id', 'i4'), ('profile_id', 'i4')])
fm_profile_dtype = np.dtype([('profile_id', 'i4'),
                             ('calcrule_id', 'i4'),
                             ('deductible_1', 'f4'),
                             ('deductible_2', 'f4'),
                             ('deductible_3', 'f4'),
                             ('attachment_1', 'f4'),
                             ('limit_1', 'f4'),
                             ('share_1', 'f4'),
                             ('share_2', 'f4'),
                             ('share_3', 'f4'),
                             ])
fm_profile_step_dtype = np.dtype([('profile_id', 'i4'),
                                  ('calcrule_id', 'i4'),
                                  ('deductible_1', 'f4'),
                                  ('deductible_2', 'f4'),
                                  ('deductible_3', 'f4'),
                                  ('attachment_1', 'f4'),
                                  ('limit_1', 'f4'),
                                  ('share_1', 'f4'),
                                  ('share_2', 'f4'),
                                  ('share_3', 'f4'),
                                  ('step_id', 'i4'),
                                  ('trigger_start', 'f4'),
                                  ('trigger_end', 'f4'),
                                  ('payout_start', 'f4'),
                                  ('payout_end', 'f4'),
                                  ('limit_2', 'f4'),
                                  ('scale_1', 'f4'),
                                  ('scale_2', 'f4'),
                                  ])
fm_profile_csv_col_map = {
    'deductible_1': 'deductible1',
    'deductible_2': 'deductible2',
    'deductible_3': 'deductible3',
    'attachment_1': 'attachment1',
    'limit_1': 'limit1',
    'share_1': 'share1',
    'share_2': 'share2',
    'share_3': 'share3',
    'limit_2': ' limit2',
    'scale_1': 'scale1',
    'scale_2': 'scale2',
}
fm_xref_dtype = np.dtype([('output_id', 'i4'), ('agg_id', 'i4'), ('layer_id', 'i4')])
fm_xref_csv_col_map = {'output_id': 'output'}

coverages_dtype = np.dtype([('coverage_id', 'i4'), ('tiv', 'f4')])

items_dtype = np.dtype([('item_id', 'i4'),
                        ('coverage_id', 'i4'),
                        ('areaperil_id', areaperil_int),
                        ('vulnerability_id', 'i4'),
                        ('group_id', 'i4')])


def load_as_ndarray(dir_path, name, _dtype, must_exist=True, col_map=None):
    """
    load a file as a numpy ndarray
    useful for multi-columns files
    Args:
        dir_path: path to the directory where the binary or csv file is stored
        name: name of the file
        _dtype: np.dtype
        must_exist: raise FileNotFoundError if no file is present
        col_map: name re-mapping to change name of csv columns
    Returns:
        numpy ndarray
    """

    if os.path.isfile(os.path.join(dir_path, name + '.bin')):
        return np.fromfile(os.path.join(dir_path, name + '.bin'), dtype=_dtype)
    elif must_exist or os.path.isfile(os.path.join(dir_path, name + '.csv')):
        # in csv column cam be out of order and have different name,
        # we load with pandas and write each column to the ndarray
        if col_map is None:
            col_map = {}
        with open(os.path.join(dir_path, name + '.csv')) as file_in:
            cvs_dtype = {col_map.get(key, key): col_dtype for key, (col_dtype, _) in _dtype.fields.items()}
            df = pd.read_csv(file_in, delimiter=',', dtype=cvs_dtype, usecols=list(cvs_dtype.keys()))
            res = np.empty(df.shape[0], dtype=_dtype)
            for name in _dtype.names:
                res[name] = df[col_map.get(name, name)]
            return res
    else:
        return np.empty(0, dtype=_dtype)


def load_as_array(dir_path, name, _dtype, must_exist=True):
    """
    load file as a single numpy array,
     useful for files with a binary version with only one type of value where their index correspond to an id.
     For example coverage.bin only contains tiv value for each coverage id
     coverage_id n correspond to index n-1
    Args:
        dir_path: path to the directory where the binary or csv file is stored
        name: name of the file
        _dtype: numpy dtype of the required array
        must_exist: raise FileNotFoundError if no file is present
    Returns:
        numpy array of dtype type
    """
    fp = os.path.join(dir_path, name + '.bin')
    if os.path.isfile(fp):
        return np.fromfile(fp, dtype=_dtype)
    elif must_exist or os.path.isfile(os.path.join(dir_path, name + '.csv')):
        fp = os.path.join(dir_path, name + '.csv')
        with open(fp) as file_in:
            return np.loadtxt(file_in, dtype=_dtype, delimiter=',', skiprows=1, usecols=1)
    else:
        return np.empty(0, dtype=_dtype)


float_equal_precision = np.finfo(oasis_float).eps


@nb.njit(cache=True)
def almost_equal(a, b):
    return abs(a - b) < float_equal_precision



__all__ = [
    "OasisBaseCommand",
    "OasisComputationCommand",
]

import logging
import os
import sys

from argparsetree import BaseCommand

from ..utils.path import PathCleaner
from ..utils.inputs import InputValues, str2bool

from ..manager import OasisManager as om


class OasisBaseCommand(BaseCommand):
    """
    The base command to inherit from for each command.

    2 additional arguments (``--verbose`` and ``--config``) are added to
    the parser so that they are available for all commands.
    """

    def __init__(self, *args, **kwargs):
        self._logger = None
        self.args = None
        self.log_verbose = False
        super(OasisBaseCommand, self).__init__(*args, **kwargs)

    def add_args(self, parser):
        """
        Adds arguments to the argument parser. This is used to modify
        which arguments are processed by the command.

        2 global parameters (``--verbose`` and ``--config``) are added
        so that they are available to all commands.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        parser.add_argument('-V', '--verbose', action='store_true', help='Use verbose logging.')
        parser.add_argument(
            '-C', '--config', required=False, type=PathCleaner('MDK config. JSON file', preexists=True),
            help='MDK config. JSON file', default='./oasislmf.json' if os.path.isfile('./oasislmf.json') else None
        )

    def parse_args(self):
        """
        Parses the command line arguments and sets them in ``self.args``

        :return: The arguments taken from the command line
        """
        try:
            self.args = super(OasisBaseCommand, self).parse_args()
            self.log_verbose = self.args.verbose
            self.setup_logger()
            return self.args
        except Exception:
            self.setup_logger()
            raise

    def setup_logger(self):
        """
        The logger to use for the command with the verbosity set
        """
        if not self._logger:
            if self.log_verbose:
                log_level = logging.DEBUG
                log_format = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
            else:
                log_level = logging.INFO
                log_format = '%(message)s'

            logger = logging.getLogger('oasislmf')
            for handler in list(logger.handlers):
                if handler.name == 'oasislmf':
                    logger.removeHandler(handler)
                    break

            ods_logger = logging.getLogger('ods_tools')
            if self.log_verbose:
                ods_logger.setLevel(logging.DEBUG)
            else:
                ods_logger.setLevel(logging.WARNING)
            ods_logger.propagate = False

            ch = logging.StreamHandler(stream=sys.stdout)
            ch.name = 'oasislmf'
            ch.setFormatter(logging.Formatter(log_format))
            logger.addHandler(ch)
            ods_logger.addHandler(ch)
            logger.setLevel(log_level)
            self._logger = logger

    @property
    def logger(self):
        if self._logger:
            return self._logger

        return logging.getLogger('oasislmf')


class OasisComputationCommand(OasisBaseCommand):
    """
    Eventually, the Parent class for all Oasis Computation Command
    create the command line interface from parameter define in the associated computation step
    """

    def add_args(self, parser):
        """
        Adds arguments to the argument parser.

        :param parser: The argument parser object
        :type parser: ArgumentParser
        """
        super().add_args(parser)

        for param in om.computations_params[self.computation_name]:
            add_argument_kwargs = {key: param.get(key) for key in ['action', 'nargs', 'const', 'type', 'choices',
                                                                   'help', 'metavar', 'dest'] if param.get(key) is not None}
            # If 'Help' is not set then this is a function only paramter, skip
            if 'help' in add_argument_kwargs:
                arg_name = f"--{param['name'].replace('_', '-')}"
                if param.get('flag'):
                    parser.add_argument(param.get('flag'), arg_name, **add_argument_kwargs)
                else:
                    parser.add_argument(arg_name, **add_argument_kwargs)

    def action(self, args):
        """
        Generic method that call the correct manager method from the child class computation_name

        :param args: The arguments from the command line
        :type args: Namespace
        """
        inputs = InputValues(args)

        _kwargs = {
            param['name']: inputs.get(param['name'], required=param.get('required'), is_path=param.get('is_path'), dtype=param.get('type')) for
            param in om.computations_params[self.computation_name]}

        # Override logger setup from kwargs
        if 'verbose' in _kwargs:
            self.logger.level = logging.DEBUG if str2bool(_kwargs.get('verbose')) else logging.INFO

        self.logger.info(f'\nStating oasislmf command - {self.computation_name}')
        manager_method = getattr(om(), om.computation_name_to_method(self.computation_name))
        manager_method(**_kwargs)



from argparse import RawDescriptionHelpFormatter
from .command import OasisBaseCommand, OasisComputationCommand


class ListApiCmd(OasisComputationCommand):
    """
    Issue API GET requests via the command line
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PlatformList'


class RunApiCmd(OasisComputationCommand):
    """
    Run a model via the Oasis Platoform API end to end
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PlatformRun'


class RunInputApiCmd(OasisComputationCommand):
    """
    Run a model via the Oasis Platoform API end to end
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PlatformRunInputs'


class RunLossApiCmd(OasisComputationCommand):
    """
    Run a model via the Oasis Platoform API end to end
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PlatformRunLosses'


class DeleteApiCmd(OasisComputationCommand):
    """
    Delete items from the Platform API
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PlatformDelete'


class GetApiCmd(OasisComputationCommand):
    """
    Download files from the Oasis Platoform API
    """
    formatter_class = RawDescriptionHelpFormatter
    computation_name = 'PlatformGet'


class ApiCmd(OasisBaseCommand):
    sub_commands = {
        'list': ListApiCmd,
        'run': RunApiCmd,
        'generate-oasis-files': RunInputApiCmd,
        'generate-losses': RunLossApiCmd,
        'delete': DeleteApiCmd,
        'get': GetApiCmd,
    }



import numpy as np

allowed_allocation_rule = [0, 1, 2, 3]

need_tiv_policy = (4, 6, 18, 27, 28, 29, 30, 31, 32, 37)
need_extras = (7, 8, 10, 11, 13, 19, 26, 27, 35, 36, 107, 108, 110, 111, 113, 119, 126, 135, 136)
EXTRA_VALUES = 3

compute_idx_dtype = np.dtype([
    ('level_start_compute_i', int),
    ('next_compute_i', int),
    ('compute_i', int),
    ('sidx_i', int),
    ('sidx_ptr_i', int),
    ('loss_ptr_i', int),
    ('extras_ptr_i', int),
])

DEDUCTIBLE = 0
OVERLIMIT = 1
UNDERLIMIT = 2



__all__ = [
    'ExposurePreAnalysis'
]

import json
import pathlib
from ods_tools.oed import UnknownColumnSaveOption

from ..base import ComputationStep
from ...utils.data import get_exposure_data, prepare_location_df
from ...utils.inputs import str2bool
from ...utils.path import get_custom_module
from ...utils.exceptions import OasisException


class ExposurePreAnalysis(ComputationStep):
    """
    Computation step that will be call before the gulcalc.
    Add the ability to specify a model specific pre-analysis hook for exposure modification,
    Allows OED to be processed by some custom code.
    Example of usage include geo-coding, exposure enhancement, or dis-aggregation...

    when the run method is call it will :
    - load the module specified at exposure_pre_analysis_module
    - init the class named exposure_pre_analysis_class_name with all the non null args in step_params as key arguments
    - call the method run of the object
    - return the output of the method

    you can find an example of such custom module in OasisPyWind/custom_module/exposure_pre_analysis.py

    """
    step_params = [{'name': 'exposure_pre_analysis_module', 'required': True, 'is_path': True, 'pre_exist': True,
                    'help': 'Exposure Pre-Analysis lookup module path'},
                   {'name': 'exposure_pre_analysis_class_name', 'default': 'ExposurePreAnalysis',
                    'help': 'Name of the class to use for the exposure_pre_analysis'},
                   {'name': 'exposure_pre_analysis_setting_json', 'is_path': True, 'pre_exist': True,
                    'help': 'Exposure Pre-Analysis config JSON file path'},
                   {'name': 'oed_schema_info', 'is_path': True, 'pre_exist': True, 'help': 'path to custom oed_schema'},
                   {'name': 'oed_location_csv', 'flag': '-x', 'is_path': True, 'pre_exist': True, 'help': 'Source location CSV file path'},
                   {'name': 'oed_accounts_csv', 'flag': '-y', 'is_path': True, 'pre_exist': True, 'help': 'Source accounts CSV file path'},
                   {'name': 'oed_info_csv', 'flag': '-i', 'is_path': True, 'pre_exist': True, 'help': 'Reinsurance info. CSV file path'},
                   {'name': 'oed_scope_csv', 'flag': '-s', 'is_path': True, 'pre_exist': True, 'help': 'Reinsurance scope CSV file path'},
                   {'name': 'check_oed', 'type': str2bool, 'const': True, 'nargs': '?', 'default': True, 'help': 'if True check input oed files'},
                   {'name': 'oasis_files_dir', 'flag': '-o', 'is_path': True, 'pre_exist': False,
                    'help': 'Path to the directory in which to generate the Oasis files'},
                   {'name': 'location', 'type': str, 'nargs': '+', 'help': 'A set of locations to include in the files'},
                   {'name': 'portfolio', 'type': str, 'nargs': '+', 'help': 'A set of portfolios to include in the files'},
                   {'name': 'account', 'type': str, 'nargs': '+', 'help': 'A set of locations to include in the files'},
                   {'name': 'base_df_engine', 'type': str, 'default': 'oasis_data_manager.df_reader.reader.OasisPandasReader',
                    'help': 'The default dataframe reading engine to use when loading files'},
                   {'name': 'exposure_df_engine', 'type': str, 'default': None,
                    'help': 'The dataframe reading engine to use when loading exposure files'},
                   {'name': 'model_df_engine', 'type': str, 'default': None, 'help': 'The dataframe reading engine to use when loading model files'},
                   {'name': 'model_data_dir', 'flag': '-d', 'is_path': True, 'pre_exist': True, 'help': 'Model data directory path'},
                   {'name': 'analysis_settings_json', 'flag': '-a', 'is_path': True, 'pre_exist': True,
                    'help': 'Analysis settings JSON file path'},
                   {'name': 'user_data_dir', 'flag': '-D', 'is_path': True, 'pre_exist': False,
                    'help': 'Directory containing additional model data files which varies between analysis runs'},
                   ]

    run_dir_key = 'pre-analysis'

    def get_exposure_data_config(self):
        return {
            'location': self.oed_location_csv,
            'account': self.oed_accounts_csv,
            'ri_info': self.oed_info_csv,
            'ri_scope': self.oed_scope_csv,
            'oed_schema_info': self.oed_schema_info,
            'check_oed': self.check_oed,
            'use_field': True,
            'location_numbers': self.location,
            'portfolio_numbers': self.portfolio,
            'account_numbers': self.account,
            'base_df_engine': self.base_df_engine,
            'exposure_df_engine': self.exposure_df_engine,
        }

    def run(self):
        """
        import exposure_pre_analysis_module and call the run method
        """
        exposure_data = get_exposure_data(self, add_internal_col=True)
        kwargs = dict()

        # If given a value for 'oasis_files_dir' then use that directly
        if self.oasis_files_dir:
            input_dir = self.oasis_files_dir
        else:
            input_dir = self.get_default_run_dir()
            pathlib.Path(input_dir).mkdir(parents=True, exist_ok=True)

        ids_option = {'loc_id': UnknownColumnSaveOption.DELETE,
                      'loc_idx': UnknownColumnSaveOption.DELETE}
        exposure_data.save(path=input_dir, version_name='raw', save_config=True, unknown_columns=ids_option)
        kwargs['exposure_data'] = exposure_data
        kwargs['input_dir'] = input_dir
        kwargs['model_data_dir'] = self.model_data_dir
        kwargs['user_data_dir'] = self.user_data_dir
        kwargs['analysis_settings_json'] = self.analysis_settings_json

        if self.exposure_pre_analysis_setting_json:
            with open(self.exposure_pre_analysis_setting_json) as exposure_pre_analysis_setting_file:
                kwargs['exposure_pre_analysis_setting'] = json.load(exposure_pre_analysis_setting_file)

        _module = get_custom_module(self.exposure_pre_analysis_module, 'Exposure Pre-Analysis lookup module path')

        try:
            _class = getattr(_module, self.exposure_pre_analysis_class_name)
        except AttributeError as e:
            raise OasisException(f"class {self.exposure_pre_analysis_class_name} "
                                 f"is not defined in module {self.exposure_pre_analysis_module}") from e.__cause__

        original_files = {oed_source.oed_name: str(oed_source.current_source['filepath']) for oed_source in exposure_data.get_oed_sources()}
        self.logger.info('\nPre-analysis original files: {}'.format(
            json.dumps(original_files, indent=4)))

        print(kwargs)
        print(_class(**kwargs))
        _class_return = _class(**kwargs).run()

        exposure_data.save(path=input_dir, version_name='', save_config=True, unknown_columns=ids_option)
        # regenerate ids
        exposure_data.location.dataframe = exposure_data.location.dataframe.drop(columns=['loc_id', 'loc_idx'])
        exposure_data.location.dataframe = prepare_location_df(exposure_data.location.dataframe)

        modified_files = {oed_source.oed_name: str(oed_source.current_source['filepath']) for oed_source in exposure_data.get_oed_sources()}
        self.logger.info('\nPre-analysis modified files: {}'.format(
            json.dumps(modified_files, indent=4)))
        return {
            "class": _class_return,
            "modified": modified_files,
            "original": original_files,
        }



import numba as nb
import numpy as np
import logging

from oasislmf.pytools.common.data import oasis_int, oasis_int_size, oasis_float, oasis_float_size
from oasislmf.pytools.common.event_stream import (EventReader, get_and_check_header_in, stream_info_to_bytes, write_mv_to_stream,
                                                  mv_read, mv_write, PIPE_CAPACITY)

logger = logging.getLogger(__name__)


@nb.jit(nopython=True, cache=True)
def read_buffer(byte_mv, cursor, valid_buff, event_id, item_id, items_amps, plafactors, default_factor, out_byte_mv, out_cursor):
    """
    read the gul loss stream, apply the post loss amplification factor and load it into out_byte_mv buffer
    This modified version of the read_buffer template return result when the whole input buffer is read and not when an event is read.
    therefore it cannot be used to read multiple stream at a time because events would be mixed up.

    Args:
        byte_mv: input byte array
        cursor: read cursor
        valid_buff: valid part of input array
        event_id: last event id
        item_id: last item_id
        items_amps (numpy array): amplification IDs where indexes correspond to item IDs
        plafactors (dict): event ID and amplification ID pairs mapped to loss factors
        default_factor (float): post loss reduction/amplification factor to be used if loss factor not found in plafactors
        out_byte_mv: output byte arrau
        out_cursor: single value array to store valid part of out_byte_mv

    """
    if item_id:
        factor = plafactors.get((event_id, items_amps[item_id]), default_factor)
    while True:
        if item_id:
            if valid_buff - cursor < (oasis_int_size + oasis_float_size):
                break
            sidx, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
            if sidx:
                loss, _ = mv_read(byte_mv, cursor, oasis_float, oasis_float_size)
                loss = 0 if np.isnan(loss) else loss

                ###### do loss read ######
                cursor = mv_write(byte_mv, cursor, oasis_float, oasis_float_size, loss * factor)
                ##########

            else:
                ##### do item exit ####
                ##########
                cursor += oasis_float_size
                item_id = 0
        else:
            if valid_buff - cursor < 2 * oasis_int_size:
                break
            event_id, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
            item_id, cursor = mv_read(byte_mv, cursor, oasis_int, oasis_int_size)
            ##### do new item setup #####
            factor = plafactors.get((event_id, items_amps[item_id]), default_factor)
            ##########
    out_byte_mv[:cursor] = byte_mv[:cursor]
    out_cursor[0] = cursor
    return cursor, event_id, item_id, 1


class PlaReader(EventReader):
    def __init__(self, items_amps, plafactors, default_factor):
        self.items_amps = items_amps
        self.plafactors = plafactors
        self.default_factor = default_factor
        self.out_byte_mv = np.empty(PIPE_CAPACITY, dtype='b')
        self.out_cursor = np.empty(1, dtype='i4')

        self.event_id = 0
        self.item_id = 0

        self.logger = logger

    def read_buffer(self, byte_mv, cursor, valid_buff, event_id, item_id):
        cursor, self.event_id, self.item_id, yield_event = read_buffer(
            byte_mv, cursor, valid_buff, self.event_id, self.item_id,
            self.items_amps, self.plafactors, self.default_factor, self.out_byte_mv, self.out_cursor
        )
        return cursor, self.event_id, self.item_id, yield_event


def read_and_write_streams(
    stream_in, stream_out, items_amps, plafactors, default_factor
):
    """
    Read input stream from gulpy or gulcalc, determine amplification ID from
    item ID, determine loss factor from event ID and amplification ID pair,
    multiply losses by relevant factors, and write to output stream.

    Input stream is binary file with layout:
        stream type (oasis_int), maximum sidx value (oasis_int),
        event ID 1 (oasis_int), item ID 1 (oasis_int),
        sample ID/sidx 1 (oasis_int), loss for sidx 1 (oasis_float),
        ...
        sample ID/sidx n (oasis_int), loss for sidx n (oasis_float),
        0 (oasis_int), 0.0 (4-byte float),
        event ID 1 (oasis_int), item ID 2 (oasis_int),
        ...
        event ID M (oasis_int), item ID N (oasis_int),
        sample ID/sidx 1 (oasis_int), loss for sidx 1 (oasis_float),
        ...
        sample ID/sidx n (oasis_int), loss for sidx n (oasis_float)

    Sample ID/sidx of 0 indicates start of next event ID-item ID pair. Output
    stream has same format as input stream.

    Args:
        stream_in (buffer): input stream
        stream_out (buffer): output stream
        items_amps (numpy array): amplification IDs where indexes correspond to item IDs
        plafactors (dict): event ID and amplification ID pairs mapped to loss factors
        default_factor (float): post loss reduction/amplification factor to be used if loss factor not found in plafactors

    """
    stream_source_type, stream_agg_type, len_sample = get_and_check_header_in(stream_in)
    stream_out.write(stream_info_to_bytes(stream_source_type, stream_agg_type))
    stream_out.write(len_sample.tobytes())

    pla_reader = PlaReader(items_amps, plafactors, default_factor)
    for _ in pla_reader.read_streams(stream_in):
        write_mv_to_stream(stream_out, pla_reader.out_byte_mv, pla_reader.out_cursor[0])
